#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""Tiff Unit tests"""
import unittest
import os
import logging

logger = logging.getLogger(__name__)

import fabio
from fabio import tifimage
from ..utilstest import UtilsTest


class TestTif(unittest.TestCase):
    # filename dim1 dim2 min max mean stddev
    TESTIMAGES = [
        ("Feb09-bright-00.300s_WAXS.bz2", 1042, 1042, 0, 65535, 8546.6414, 1500.4198),
        ("Feb09-bright-00.300s_WAXS.gz", 1042, 1042, 0, 65535, 8546.6414, 1500.4198),
        ("Feb09-bright-00.300s_WAXS", 1042, 1042, 0, 65535, 8546.6414, 1500.4198)]

    def test_read(self):
        """
        Test the reading of Mar345 images
        """
        for params in self.TESTIMAGES:
            name = params[0]
            logger.debug("Processing: %s" % name)
            dim1, dim2 = params[1:3]
            shape = dim2, dim1
            mini, maxi, mean, stddev = params[3:]
            obj = fabio.tifimage.TifImage()
            obj.read(UtilsTest.getimage(name))

            self.assertAlmostEqual(mini, obj.getmin(), 2, "getmin [%s,%s]" % (mini, obj.getmin()))
            self.assertAlmostEqual(maxi, obj.getmax(), 2, "getmax [%s,%s]" % (maxi, obj.getmax()))
            self.assertAlmostEqual(mean, obj.getmean(), 2, "getmean [%s,%s]" % (mean, obj.getmean()))
            self.assertAlmostEqual(stddev, obj.getstddev(), 2, "getstddev [%s,%s]" % (stddev, obj.getstddev()))
            self.assertEqual(shape, obj.shape, "dim1")

    def test_header(self):
        for params in self.TESTIMAGES:
            name = params[0]
            logger.debug("Processing: %s" % name)
            obj = fabio.tifimage.TifImage()
            obj.read(UtilsTest.getimage(name))

            # The key order is not the same depending on Python2 or 3
            expected_keys = set([
                'info',
                'photometricInterpretation',
                'rowsPerStrip',
                'nColumns',
                'compression',
                'sampleFormat',
                'imageDescription',
                'nRows',
                'colormap',
                'nBits',
                'date',
                'software',
                'compression_type',
                'stripOffsets',
                'stripByteCounts'])
            self.assertEqual(set(obj.header.keys()), expected_keys)

    def test_frame(self):
        for params in self.TESTIMAGES:
            name = params[0]
            logger.debug("Processing: %s" % name)
            dim1, dim2 = params[1:3]
            obj = fabio.tifimage.TifImage()
            obj.read(UtilsTest.getimage(name))

            self.assertEqual(obj.nframes, 1)
            frame = obj.getframe(0)
            self.assertIsNotNone(frame)
            self.assertIsNotNone(frame.data)
            self.assertEqual(frame.data.shape, (dim2, dim1))
            self.assertEqual(len(frame.header.keys()), 15)


class TestTifImage_Pilatus(unittest.TestCase):

    def setUp(self):
        self.fn = {}
        for i in ["pilatus2M.tif", "pilatus2M.edf"]:
            self.fn[i] = UtilsTest.getimage(i + ".bz2")
        for i in self.fn:
            assert os.path.exists(self.fn[i])

    def test1(self):
        """
        Testing pilatus tif bug
        """
        o1 = fabio.open(self.fn["pilatus2M.tif"]).data
        o2 = fabio.open(self.fn["pilatus2M.edf"]).data
        self.assertEqual(abs(o1 - o2).max(), 0.0)


class TestTifImage_Packbits(unittest.TestCase):

    def setUp(self):
        self.fn = {}
        for i in ["oPPA_5grains_0001.tif", "oPPA_5grains_0001.edf"]:
            self.fn[i] = UtilsTest.getimage(i + ".bz2")
        for i in self.fn:
            assert os.path.exists(self.fn[i])

    def test1(self):
        """
        Testing packbit comressed data tif bug
        """
        o1 = fabio.open(self.fn["oPPA_5grains_0001.tif"]).data
        o2 = fabio.open(self.fn["oPPA_5grains_0001.edf"]).data
        self.assertEqual(abs(o1 - o2).max(), 0.0)


class TestTifImage_fit2d(unittest.TestCase):

    def setUp(self):
        self.fn = {}
        for i in ["fit2d.tif", "fit2d.edf"]:
            self.fn[i] = UtilsTest.getimage(i + ".bz2")
        for i in self.fn:
            assert os.path.exists(self.fn[i])

    def test1(self):
        """
        Testing packbit comressed data tif bug
        """
        o1 = fabio.open(self.fn["fit2d.tif"]).data
        o2 = fabio.open(self.fn["fit2d.edf"]).data
        self.assertEqual(abs(o1 - o2).max(), 0.0)


class TestTifImage_A0009(unittest.TestCase):
    """
    test image from ??? with this error
a0009.tif TIFF 1024x1024 1024x1024+0+0 16-bit Grayscale DirectClass 2MiB 0.000u 0:00.010
identify: a0009.tif: invalid TIFF directory; tags are not sorted in ascending order. `TIFFReadDirectory' @ tiff.c/TIFFWarnings/703.
identify: a0009.tif: TIFF directory is missing required "StripByteCounts" field, calculating from imagelength. `TIFFReadDirectory' @ tiff.c/TIFFWarnings/703.

    """

    def setUp(self):
        self.fn = {}
        for i in ["a0009.tif", "a0009.edf"]:
            self.fn[i] = UtilsTest.getimage(i + ".bz2")[:-4]
        for i in self.fn:
            assert os.path.exists(self.fn[i])

    def test1(self):
        """
        Testing packbit comressed data tif bug
        """
        o1 = fabio.open(self.fn["a0009.tif"]).data
        o2 = fabio.open(self.fn["a0009.edf"]).data
        self.assertEqual(abs(o1 - o2).max(), 0.0)


class TestGzipTif(unittest.TestCase):

    def setUp(self):
        self.unzipped = UtilsTest.getimage("oPPA_5grains_0001.tif.bz2")[:-4]
        self.zipped = self.unzipped + ".gz"
        assert os.path.exists(self.zipped)
        assert os.path.exists(self.unzipped)

    def test1(self):
        o1 = fabio.open(self.zipped)
        o2 = fabio.open(self.unzipped)
        self.assertEqual(o1.data[0, 0], 10)
        self.assertEqual(o2.data[0, 0], 10)


class TestTif_Rect(unittest.TestCase):

    def setUp(self):
        self.fn = UtilsTest.getimage("testmap1_0002.tif.bz2")[:-4]

    def test1(self):
        for ext in ["", ".gz", ".bz2"]:
            o1 = fabio.open(self.fn + ext)
            self.assertEqual(o1.data.shape, (100, 120))


class TestTif_Colormap(unittest.TestCase):

    def setUp(self):
        self.fn = UtilsTest.getimage("indexed_color.tif.bz2")[:-4]

    def tearDown(self):
        tifimage._USE_PIL = True
        tifimage._USE_TIFFIO = True

    def _test_base(self):
        for ext in ["", ".gz", ".bz2"]:
            image = fabio.open(self.fn + ext)
            self.assertEqual(image.data.shape, (16, 16, 3))
            self.assertEqual(image.data[0, 0].tolist(), [255, 0, 0])
            self.assertEqual(image.data[8, 8].tolist(), [0, 252, 255])
            self.assertEqual(image.data[15, 15].tolist(), [255, 96, 0])

    def test_pil(self):
        if tifimage.PIL is None:
            self.skipTest("PIL is not available")
        tifimage._USE_TIFFIO = False
        self._test_base()

    def test_tiffio(self):
        tifimage._USE_PIL = False
        self._test_base()


class TestTif_LibTiffPic(unittest.TestCase):
    """Test TIFF loading from libtiffpic dataset"""

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.files = UtilsTest.resources.getdir("libtiffpic.tar.gz")

    def test_all_images(self):
        if tifimage.PIL is None:
            self.skipTest("PIL is not available")
        for filename in self.files:
            if filename.endswith(".tif"):
                with fabio.open(filename):
                    pass


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestTif))
    testsuite.addTest(loadTests(TestGzipTif))
    testsuite.addTest(loadTests(TestTif_Rect))
    testsuite.addTest(loadTests(TestTifImage_A0009))
    testsuite.addTest(loadTests(TestTifImage_fit2d))
    testsuite.addTest(loadTests(TestTifImage_Packbits))
    testsuite.addTest(loadTests(TestTifImage_Pilatus))
    testsuite.addTest(loadTests(TestTif_Colormap))
    testsuite.addTest(loadTests(TestTif_LibTiffPic))
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""
Test JPEG format
"""

import unittest
import os
import shutil
import logging

logger = logging.getLogger(__name__)

import fabio
from ... import jpegimage
from ..utilstest import UtilsTest

TEST_DIRECTORY = None
# Temporary directory where storing test data


def setUpModule():
    global TEST_DIRECTORY
    TEST_DIRECTORY = os.path.join(UtilsTest.tempdir, __name__)
    os.makedirs(TEST_DIRECTORY)


def tearDownModule():
    shutil.rmtree(TEST_DIRECTORY)


class TestJpegImage(unittest.TestCase):
    """Test the class format"""

    def setUp(self):
        if jpegimage.Image is None:
            self.skipTest("PIL is not available")

    def test_read_uint8(self):
        filename = UtilsTest.getimage("rand_uint8.jpg.bz2")[:-4]
        image_format = jpegimage.JpegImage()
        image = image_format.read(filename)
        self.assertEqual(image.data.shape, (64, 64))
        self.assertIn("jfif", image.header)

    def test_read_failing_file(self):
        filename = os.path.join(TEST_DIRECTORY, "2.jpg")
        filename_source = UtilsTest.getimage("rand_uint8.jpg.bz2")[:-4]

        with open(filename_source, "r+b") as fsource:
            with open(filename, "w+b") as ftest:
                ftest.write(fsource.read())
                ftest.seek(1)
                ftest.write(b".")

        image_format = jpegimage.JpegImage()
        try:
            _image = image_format.read(filename)
            self.fail()
        except IOError:
            pass

    def test_read_empty_file(self):
        filename = os.path.join(TEST_DIRECTORY, "3.jpg")
        f = open(filename, "wb")
        f.close()

        image_format = jpegimage.JpegImage()
        try:
            _image = image_format.read(filename)
            self.fail()
        except IOError:
            pass

    def test_read_missing_file(self):
        filename = os.path.join(TEST_DIRECTORY, "4.jpg")

        image_format = jpegimage.JpegImage()
        try:
            _image = image_format.read(filename)
            self.fail()
        except IOError:
            pass


class TestPilNotAvailable(unittest.TestCase):

    def setUp(self):
        filename = UtilsTest.getimage("rand_uint8.jpg.bz2")[:-4]
        self.filename = filename

        self.old = jpegimage.Image

    def tearDown(self):
        jpegimage.Image = self.old
        self.filename = None
        self.data = None

    def open_image(self):
        return fabio.open(self.filename)

    def test_with_pil(self):
        if jpegimage.Image is None:
            self.skipTest("PIL is not available")
        image = self.open_image()
        self.assertIsInstance(image, jpegimage.JpegImage)
        self.assertEqual(image.data.shape, (64, 64))
        self.assertIn("jfif", image.header)

    def test_without_pil(self):
        try:
            old = jpegimage.Image
            jpegimage.Image = None
            try:
                _image = self.open_image()
                self.fail()
            except IOError:
                pass
        finally:
            jpegimage.Image = old


class TestJpegImageInsideFabio(unittest.TestCase):
    """Test the format inside the fabio framework"""

    def test_read_uint8(self):
        if jpegimage.Image is None:
            self.skipTest("PIL is not available")
        filename = UtilsTest.getimage("rand_uint8.jpg.bz2")[:-4]
        image = fabio.open(filename)
        self.assertIsInstance(image, jpegimage.JpegImage)
        self.assertEqual(image.data.shape, (64, 64))
        self.assertIn("jfif", image.header)


def suite():
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loader(TestJpegImage))
    testsuite.addTest(loader(TestJpegImageInsideFabio))
    testsuite.addTest(loader(TestPilNotAvailable))
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
"""
#bruker100 Unit tests

19/01/2015
"""

import unittest
import os
import logging

logger = logging.getLogger(__name__)

import numpy
from fabio.bruker100image import Bruker100Image, _split_data, _merge_data
from fabio.openimage import openimage
from ..utilstest import UtilsTest

# filename dim1 dim2 min max mean stddev
TESTIMAGES = """NaCl_10_01_0009.sfrm         512 512 -30 5912 34.4626 26.189
                NaCl_10_01_0009.sfrm.gz      512 512 -30 5912 34.4626 26.189
                NaCl_10_01_0009.sfrm.bz2     512 512 -30 5912 34.4626 26.189"""
REFIMAGE = "NaCl_10_01_0009.npy.bz2"


class TestBruker100(unittest.TestCase):
    """ check some read data from bruker version100 detector"""

    def setUp(self):
        """
        download images
        """
        UtilsTest.getimage(REFIMAGE)
        self.im_dir = os.path.dirname(UtilsTest.getimage(TESTIMAGES.split()[0] + ".bz2"))

    def test_read(self):
        """ check we can read bruker100 images"""
        for line in TESTIMAGES.split("\n"):
            vals = line.split()
            name = vals[0]
            dim1, dim2 = [int(x) for x in vals[1:3]]
            shape = dim2, dim1
            mini, maxi, mean, stddev = [float(x) for x in vals[3:]]
            obj = Bruker100Image()
            obj.read(os.path.join(self.im_dir, name))
            self.assertAlmostEqual(mini, obj.getmin(), 2, "getmin")
            self.assertAlmostEqual(maxi, obj.getmax(), 2, "getmax")
            self.assertAlmostEqual(mean, obj.getmean(), 2, "getmean")
            self.assertAlmostEqual(stddev, obj.getstddev(), 2, "getstddev")
            self.assertEqual(shape, obj.shape)

    def test_same(self):
        """ check we can read bruker100 images"""
        ref = openimage(os.path.join(self.im_dir, REFIMAGE))
        for line in TESTIMAGES.split("\n"):
            obt = openimage(os.path.join(self.im_dir, line.split()[0]))
            self.assertTrue(abs(ref.data - obt.data).max() == 0, "data are the same")

    def test_write(self):
        fname = TESTIMAGES.split()[0]
        obt = openimage(os.path.join(self.im_dir, fname))
        name = os.path.basename(fname)

        obj = Bruker100Image(data=obt.data, header=obt.header)
        obj.write(os.path.join(UtilsTest.tempdir, name))
        other = openimage(os.path.join(UtilsTest.tempdir, name))
        self.assertEqual(abs(obt.data - other.data).max(), 0, "data are the same")
        for key in obt.header:
            self.assertTrue(key in other.header, "Key %s is in header" % key)
            self.assertEqual(obt.header[key], other.header[key], "value are the same for key %s" % key)
        os.unlink(os.path.join(UtilsTest.tempdir, name))

    def test_split_merge(self):
        "Pretty challanging random example"
        shape = 256, 256
        outliers = 100
        a = numpy.random.normal(100, 50, size=shape)
        a.ravel()[numpy.random.randint(0, numpy.prod(shape), size=outliers)] = numpy.random.randint(10000, 1000000, size=outliers)
        ref = a.astype("int32")

        for baseline in (None, 0, False):
            split = _split_data(ref, baseline=baseline)
            logger.info("size of underflow: %s overflow1 %s overflow2: %s",
                        split["underflow"].shape, split["overflow1"].shape, split["overflow2"].shape)
            obt = _merge_data(**split)
            self.assertTrue(numpy.allclose(obt, ref), f"data are the same, baseline={baseline}")

    def test_conversion(self):
        fname = UtilsTest.getimage("testcbf.cbf.bz2")[:-4]
        c = openimage(fname)
        assert "Cbf" in c.__class__.__name__, "This is a CbfImage"
        b = c.convert("bruker100")
        fname_out = os.path.join(UtilsTest.tempdir, "testcbf2bruker100.sfrm")
        b.write(fname_out)
        a = openimage(fname_out)
        self.assertTrue(numpy.allclose(a.data, c.data), msg="data are the same")


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestBruker100))
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
test cases for fileseries

28/11/2014
"""

import unittest
import logging
import os
import shutil
import numpy

logger = logging.getLogger(__name__)

import fabio
from fabio.file_series import numbered_file_series, file_series, filename_series
from fabio.file_series import FileSeries
from .utilstest import UtilsTest


class TestRandomSeries(unittest.TestCase):
    """arbitrary series"""

    def setUp(self):
        """sets up"""
        self.fso = file_series(["first", "second", "last"])

    def testfirst(self):
        """check first"""
        self.assertEqual("first", self.fso.first())

    def testlast(self):
        """check first"""
        self.assertEqual("last", self.fso.last())

    def testjump(self):
        """check jump"""
        self.assertEqual("second", self.fso.jump(1))


class TestEdfNumbered(unittest.TestCase):
    """
    Typical sequence of edf files
    """

    def setUp(self):
        """ note extension has the . in it"""
        self.fso = numbered_file_series("mydata", 0, 10005, ".edf")

    def testfirst(self):
        """ first in series"""
        self.assertEqual(self.fso.first(), "mydata0000.edf")

    def testlast(self):
        """ last in series"""
        self.assertEqual(self.fso.last(), "mydata10005.edf")

    def testnext(self):
        """ check all in order """
        mylist = ["mydata%04d.edf" % (i) for i in range(0, 10005)]
        i = 1
        while i < len(mylist):
            self.assertEqual(mylist[i], self.fso.next())
            i += 1

    def testprevious(self):
        """ check all in order """
        mylist = ["mydata%04d.edf" % (i) for i in range(0, 10005)]
        i = 10003
        self.fso.jump(10004)
        while i > 0:
            self.assertEqual(mylist[i], self.fso.previous())
            i -= 1

    def testprevjump(self):
        """check current"""
        self.fso.jump(9999)
        self.assertEqual("mydata9999.edf", self.fso.current())
        self.assertEqual("mydata9998.edf", self.fso.previous())

    def testnextjump(self):
        """check current"""
        self.fso.jump(9999)
        self.assertEqual("mydata9999.edf", self.fso.current())
        self.assertEqual("mydata10000.edf", self.fso.next())

    def testlen(self):
        """check len"""
        self.assertEqual(self.fso.len(), 10006)  # +1 for 0000


class TestFileSeries(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.tmp_directory = os.path.join(UtilsTest.tempdir, cls.__name__)
        os.makedirs(cls.tmp_directory)
        cls.create_resources()

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmp_directory)

    @classmethod
    def create_resources(cls):
        data = (numpy.random.rand(8, 8) * 10).astype(int)
        data0 = numpy.zeros((8, 8), dtype=int)
        data1 = numpy.zeros((8, 8), dtype=int) + 1
        data2 = numpy.zeros((8, 8), dtype=int) + 2

        # Single frame
        cls.create_edf_file("image_a_000.edf", [data0])
        cls.create_edf_file("image_a_001.edf", [data1])
        cls.create_edf_file("image_a_002.edf", [data2])

        # Many frames
        cls.create_edf_file("image_b_000.edf", [data0, data, data])
        cls.create_edf_file("image_b_001.edf", [data, data, data1])
        cls.create_edf_file("image_b_002.edf", [data2])

        # Any frames
        cls.create_edf_file("image_c_000.edf", [data0, data, data])
        cls.create_edf_file("image_c_001.edf", [data, data1])
        cls.create_edf_file("image_c_002.edf", [data, data, data, data])
        cls.create_edf_file("image_c_003.edf", [data2])

    @classmethod
    def get_filename(cls, filename):
        return os.path.join(cls.tmp_directory, filename)

    @classmethod
    def create_edf_file(cls, filename, data_list):
        filename = cls.get_filename(filename)
        image = fabio.factory("edfimage")
        for frame_id, data in enumerate(data_list):
            header = {"frame_id": "%d" % frame_id, "filename": "%s" % filename}
            if frame_id == 0:
                image.data = data
                image.header.update(header)
            else:
                image.append_frame(data=data, header=header)
        image.write(filename)

    def get_singleframe_files(self):
        filenames = ["image_a_000.edf", "image_a_001.edf", "image_a_002.edf"]
        filenames = [self.get_filename(f) for f in filenames]
        return filenames

    def get_multiframe_files(self):
        filenames = ["image_b_000.edf", "image_b_001.edf", "image_b_002.edf"]
        filenames = [self.get_filename(f) for f in filenames]
        return filenames

    def get_anyframe_files(self):
        filenames = ["image_c_000.edf", "image_c_001.edf", "image_c_002.edf", "image_c_003.edf"]
        filenames = [self.get_filename(f) for f in filenames]
        return filenames

    def test_singleframe_nofiles(self):
        serie = FileSeries(filenames=[], single_frame=True)
        self.assertEqual(serie.nframes, 0)
        for _ in serie.frames():
            self.fail()
        self.assertRaises(IndexError, serie.get_frame, 0)
        serie.close()

    def test_singleframe_nframes(self):
        filenames = self.get_singleframe_files()
        serie = FileSeries(filenames=filenames, single_frame=True)
        self.assertTrue(serie.nframes, 3)
        serie.close()

    def test_singleframe_nframes2(self):
        filenames = self.get_singleframe_files()
        serie = FileSeries(filenames=filenames, fixed_frames=True)
        self.assertTrue(serie.nframes, 3)
        serie.close()

    def test_singleframe_nframes3(self):
        filenames = self.get_singleframe_files()
        serie = FileSeries(filenames=filenames, fixed_frame_number=1)
        self.assertTrue(serie.nframes, 3)
        serie.close()

    def test_singleframe_getframe(self):
        filenames = self.get_singleframe_files()
        serie = FileSeries(filenames=filenames, single_frame=True)
        self.assertRaises(IndexError, serie.get_frame, -1)
        self.assertEqual(serie.get_frame(0).data[0, 0], 0)
        self.assertEqual(serie.get_frame(1).data[0, 0], 1)
        self.assertEqual(serie.get_frame(2).data[0, 0], 2)
        self.assertRaises(IndexError, serie.get_frame, 4)
        serie.close()

    def test_singleframe_frames(self):
        filenames = self.get_singleframe_files()
        serie = FileSeries(filenames=filenames, single_frame=True)
        for frame_id, frame in enumerate(serie.frames()):
            self.assertEqual(frame.data[0, 0], frame_id)
            self.assertEqual(frame.header["frame_id"], "0")
            self.assertIn("%03d" % frame_id, frame.header["filename"])
        self.assertEqual(frame_id, 2)
        serie.close()

    def test_multiframe_nofiles(self):
        serie = FileSeries(filenames=[], fixed_frames=True)
        self.assertEqual(serie.nframes, 0)
        for _ in serie.frames():
            self.fail()
        self.assertRaises(IndexError, serie.get_frame, 0)
        serie.close()

    def test_multiframe_nframes(self):
        filenames = self.get_multiframe_files()
        serie = FileSeries(filenames=filenames, fixed_frames=True)
        self.assertTrue(serie.nframes, 7)
        serie.close()

    def test_multiframe_getframe(self):
        filenames = self.get_multiframe_files()
        serie = FileSeries(filenames=filenames, fixed_frames=True)
        self.assertRaises(IndexError, serie.get_frame, -1)
        self.assertEqual(serie.get_frame(0).data[0, 0], 0)
        self.assertEqual(serie.get_frame(5).data[0, 0], 1)
        self.assertEqual(serie.get_frame(6).data[0, 0], 2)
        self.assertRaises(IndexError, serie.get_frame, 7)
        serie.close()

    def test_multiframe_frames(self):
        filenames = self.get_multiframe_files()
        serie = FileSeries(filenames=filenames, fixed_frames=True)
        for frame_id, frame in enumerate(serie.frames()):
            if frame_id not in [0, 5, 6]:
                continue
            expected_frame_id = {0: 0, 5: 2, 6: 0}[frame_id]
            expected_file_num = {0: 0, 5: 1, 6: 2}[frame_id]
            expected_data = {0: 0, 5: 1, 6: 2}[frame_id]

            self.assertEqual(frame.data[0, 0], expected_data)
            self.assertEqual(frame.header["frame_id"], "%d" % expected_frame_id)
            self.assertIn("%03d" % expected_file_num, frame.header["filename"])
        self.assertEqual(frame_id, 6)
        serie.close()

    def test_anyframe_nofiles(self):
        serie = FileSeries(filenames=[], fixed_frames=False)
        self.assertEqual(serie.nframes, 0)
        for _ in serie.frames():
            self.fail()
        self.assertRaises(IndexError, serie.get_frame, 0)
        serie.close()

    def test_anyframe_nframes(self):
        filenames = self.get_anyframe_files()
        serie = FileSeries(filenames=filenames, fixed_frames=False)
        self.assertTrue(serie.nframes, 10)
        serie.close()

    def test_anyframe_getframe(self):
        filenames = self.get_anyframe_files()
        serie = FileSeries(filenames=filenames, fixed_frames=False)
        self.assertRaises(IndexError, serie.get_frame, -1)
        self.assertEqual(serie.get_frame(0).data[0, 0], 0)
        # Reach frame 3 to force frame 4 to come from the file description cache
        serie.get_frame(3)
        self.assertEqual(serie.get_frame(4).data[0, 0], 1)
        self.assertEqual(serie.get_frame(9).data[0, 0], 2)
        self.assertRaises(IndexError, serie.get_frame, 10)
        serie.close()

    def test_anyframe_getframes(self):
        filenames = self.get_anyframe_files()
        serie = FileSeries(filenames=filenames, fixed_frames=False)
        for frame_id in range(serie.nframes):
            frame = serie.get_frame(frame_id)
            if frame_id not in [0, 4, 9]:
                continue
            expected_file_frame_id = {0: 0, 4: 1, 9: 0}[frame_id]
            expected_file_num = {0: 0, 4: 1, 9: 3}[frame_id]
            expected_data = {0: 0, 4: 1, 9: 2}[frame_id]
            expected_filename = filenames[expected_file_num]

            self.assertEqual(frame.data[0, 0], expected_data)
            self.assertEqual(frame.header["frame_id"], "%d" % expected_file_frame_id)
            self.assertEqual(frame.index, frame_id)
            self.assertEqual(frame.file_index, expected_file_frame_id)
            self.assertIs(frame.container, serie)
            self.assertIs(frame.file_container.filename, expected_filename)
            self.assertIn("%03d" % expected_file_num, frame.header["filename"])
        self.assertEqual(frame_id, 9)
        serie.close()

    def test_anyframe_frames(self):
        filenames = self.get_anyframe_files()
        serie = FileSeries(filenames=filenames, fixed_frames=False)
        for frame_id, frame in enumerate(serie.frames()):
            if frame_id not in [0, 4, 9]:
                continue
            expected_file_frame_id = {0: 0, 4: 1, 9: 0}[frame_id]
            expected_file_num = {0: 0, 4: 1, 9: 3}[frame_id]
            expected_data = {0: 0, 4: 1, 9: 2}[frame_id]
            expected_filename = filenames[expected_file_num]

            self.assertEqual(frame.data[0, 0], expected_data)
            self.assertEqual(frame.header["frame_id"], "%d" % expected_file_frame_id)
            self.assertEqual(frame.index, frame_id)
            self.assertEqual(frame.file_index, expected_file_frame_id)
            self.assertIs(frame.container, serie)
            self.assertIs(frame.file_container.filename, expected_filename)
            self.assertIn("%03d" % expected_file_num, frame.header["filename"])
        self.assertEqual(frame_id, 9)
        serie.close()

    def test_filename_iterator(self):
        filenames = self.get_anyframe_files()
        serie = FileSeries(filenames=iter(filenames))
        self.assertEqual(serie.nframes, 10)
        serie.close()

    def test_filename_generator(self):

        def generator():
            filenames = self.get_anyframe_files()
            for filename in filenames:
                yield filename

        serie = FileSeries(filenames=generator())
        self.assertEqual(serie.nframes, 10)
        serie.close()

    def test_with_numbered_file_series(self):
        filenames = numbered_file_series(self.get_filename("image_c_"), 0, 3, ".edf", digits=3)
        serie = FileSeries(filenames=filenames)
        self.assertEqual(serie.nframes, 10)
        serie.close()

    def test_with_filename_series(self):
        first_filename = self.get_filename("image_c_000.edf")
        filenames = filename_series(first_filename)
        serie = FileSeries(filenames=filenames)
        self.assertEqual(serie.nframes, 10)
        serie.close()


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestRandomSeries))
    testsuite.addTest(loadTests(TestEdfNumbered))
    testsuite.addTest(loadTests(TestFileSeries))
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE

"""

Authors:
........
* Henning O. Sorensen & Erik Knudsen:
  Center for Fundamental Research: Metal Structures in Four Dimensions;
  Risoe National Laboratory;
  Frederiksborgvej 399;
  DK-4000 Roskilde;
  email:erik.knudsen@risoe.dk
* Jon Wright, Jérôme Kieffer & Gaël Goret:
  European Synchrotron Radiation Facility;
  Grenoble (France)

Supports Mar345 imaging plate and Mar555 flat panel

Documentation on the format is available from:
http://rayonix.com/site_media/downloads/mar345_formats.pdf
"""

__authors__ = ["Henning O. Sorensen", "Erik Knudsen", "Jon Wright", "Jérôme Kieffer"]
__date__ = "03/04/2020"
__status__ = "production"
__copyright__ = "2007-2009 Risoe National Laboratory; 2010-2020 ESRF"
__licence__ = "MIT"

import struct
import time
import logging
import numpy

import fabio
from .fabioimage import FabioImage

logger = logging.getLogger(__name__)
from .compression import compPCK, decPCK


class Mar345Image(FabioImage):
    _need_a_real_file = True

    DESCRIPTION = "File format from Mar345 imaging plate and Mar555 flat panel"

    DEFAULT_EXTENSIONS = ["mar2300", "mar1200", "mar1600", "mar2000",  # 150µm pixel size
                          "mar3450", "mar3000", "mar2400", "mar1800"]  # 100µm pixel size

    def __init__(self, *args, **kwargs):
        FabioImage.__init__(self, *args, **kwargs)
        self.numhigh = None
        self.numpixels = None
        self.swap_needed = None

    def read(self, fname, frame=None):
        """ Read a mar345 image"""
        self.filename = fname
        f = self._open(self.filename, "rb")
        self._readheader(f)
        if 'compressed' in self.header['Format']:
            dim2, dim1 = self._shape
            self.data = decPCK(f, dim1, dim2, self.numhigh, swap_needed=self.swap_needed)
            self._shape = None
        else:
            logger.error("Cannot handle these formats yet due to lack of documentation")
            return None
        self._bytecode = numpy.uint32
        f.close()
        return self

    def _readheader(self, infile=None):
        """ Read a mar345 image header """
        # clip was not used anywhere - commented out
        # clip = '\x00'
        # using a couple of local variables inside this function
        f = infile
        h = {}

        # header is 4096 bytes long
        data = f.read(64)
        # the contents of the mar345 header is taken to be as
        # described in
        # http://www.mar-usa.com/support/downloads/mar345_formats.pdf
        # the first 64 bytes are 4-byte integers (but in the CBFlib
        # example image it seems to 128 bytes?)
        # first 4-byte integer is a marker to check endianness
        if struct.unpack("<i", data[0:4])[0] == 1234:
            fs = '<i'
            self.swap_needed = not(numpy.little_endian)
            logger.debug("Going for little endian, swap_needed %s" % self.swap_needed)
        else:
            self.swap_needed = numpy.little_endian
            fs = '>i'
            logger.debug("Going for big endian, swap_needed %s" % self.swap_needed)

        # image dimensions
        dim1 = int(struct.unpack(fs, data[4:8])[0])
        # number of high intensity pixels
        self.numhigh = struct.unpack(fs, data[2 * 4: (2 + 1) * 4])[0]
        h['NumHigh'] = self.numhigh
        # Image format
        i = struct.unpack(fs, data[3 * 4: (3 + 1) * 4])[0]
        if i == 1:
            h['Format'] = 'compressed'
        elif i == 2:
            h['Format'] = 'spiral'
        else:
            h['Format'] = 'compressed'
            logger.warning("image format could not be determined" +
                           "- assuming compressed mar345")
        # collection mode
        h['Mode'] = {0: 'Dose', 1: 'Time'}[struct.unpack(fs, data[4 * 4:(4 + 1) * 4])[0]]
        # total number of pixels
        self.numpixels = struct.unpack(fs, data[5 * 4:(5 + 1) * 4])[0]
        h['NumPixels'] = str(self.numpixels)
        dim2 = self.numpixels // dim1
        self._shape = dim2, dim1
        # pixel dimensions (length,height) in mm
        h['PixelLength'] = struct.unpack(fs, data[6 * 4:(6 + 1) * 4])[0] / 1000.0
        h['PixelHeight'] = struct.unpack(fs, data[7 * 4:(7 + 1) * 4])[0] / 1000.0
        # x-ray wavelength in AA
        h['Wavelength'] = struct.unpack(fs, data[8 * 4:(8 + 1) * 4])[0] / 1000000.0
        # used distance
        h['Distance'] = struct.unpack(fs, data[9 * 4:(9 + 1) * 4])[0] / 1000.0
        # starting and ending phi
        h['StartPhi'] = struct.unpack(fs, data[10 * 4:11 * 4])[0] / 1000.0
        h['EndPhi'] = struct.unpack(fs, data[11 * 4:12 * 4])[0] / 1000.0
        # starting and ending omega
        h['StartOmega'] = struct.unpack(fs, data[12 * 4:13 * 4])[0] / 1000.0
        h['EndOmega'] = struct.unpack(fs, data[13 * 4:14 * 4])[0] / 1000.0
        # Chi and Twotheta angles
        h['Chi'] = struct.unpack(fs, data[14 * 4:15 * 4])[0] / 1000.0
        h['TwoTheta'] = struct.unpack(fs, data[15 * 4:16 * 4])[0] / 1000.0

        # the rest of the header is ascii
        # TODO: validate these values against the binaries already read
        data = f.read(128)
        if b'mar research' not in data:
            logger.warning("the string \"mar research\" should be in " +
                           "bytes 65-76 of the header but was not")
            start = 128
        else:
            start = data.index(b'mar research')
            f.seek(64 + start)
        data = f.read(4096 - start - 64).strip()
        for m in data.splitlines():
            try:
                m = m.decode("ASCII")
            except UnicodeDecodeError:
                if m.startswith(b"DATE"):
                    m = m[:39].decode("ASCII")
                else:
                    logger.warning("Skip binary trash on header line %s" % m)
                    continue
            if m == 'END OF HEADER':
                break
            n = m.split(' ', 1)
            if n[0] == '':
                continue
            if n[0] in ('PROGRAM', 'DATE', 'SCANNER', 'HIGH', 'MULTIPLIER',
                        'GAIN', 'WAVELENGTH', 'DISTANCE', 'RESOLUTION',
                        'CHI', 'TWOTHETA', 'MODE', 'TIME', 'GENERATOR',
                        'MONOCHROMATOR', 'REMARK'):
                logger.debug("reading: %s %s", n[0], n[1])
                h[n[0]] = n[1].strip()
                continue
            if n[0] in ('FORMAT'):
                (h['DIM'], h['FORMAT_TYPE'], h['NO_PIXELS']) = n[1].split()
                continue
            if n[0] in ('PIXEL', 'OFFSET', 'PHI', 'OMEGA', 'COUNTS',
                        'CENTER', 'INTENSITY', 'HISTOGRAM', 'COLLIMATOR'):
                n = m.split()
                h.update([(n[0] + '_' + n[j], n[j + 1]) for j in range(1, len(n), 2)])
                continue
        self.header = h
        return h

    def write(self, fname):
        """Try to write mar345 file.
        It uses a MIT implementation of the CCP4 (LGPL) PCK1 algo from JPA"""
        bin_headers = self.binary_header()
        asc_headers = self.ascii_header("\n", 4096 - len(bin_headers)).encode("ASCII")
        hotpixels = self._high_intensity_pixel_records()
        compressed_stream = compPCK(self.data)
        with self._open(fname, mode="wb") as outfile:
            outfile.write(bin_headers)
            outfile.write(asc_headers)
            outfile.write(hotpixels)
            outfile.write(compressed_stream)
            outfile.close()

    def binary_header(self):
        """
        :return: Binary header of mar345 file
        """
        dim2, dim1 = self.shape
        self.header["HIGH"] = str(self.nb_overflow_pixels())
        binheader = numpy.zeros(16, "int32")
        binheader[0] = 1234
        binheader[1] = dim1
        binheader[2] = self.nb_overflow_pixels()
        binheader[3] = 1
        binheader[4] = (self.header.get("MODE", "TIME") == "TIME")
        binheader[5] = dim1 * dim2
        binheader[6] = int(self.header.get("PIXEL_LENGTH", 1))
        binheader[7] = int(self.header.get("PIXEL_HEIGHT", 1))
        binheader[8] = int(float(self.header.get("WAVELENGTH", 1)) * 1e6)
        binheader[9] = int(float(self.header.get("DISTANCE", 1)) * 1e3)
        binheader[10] = int(float(self.header.get("PHI_START", 1)) * 1e3)
        binheader[11] = int(float(self.header.get("PHI_END", 1)) * 1e3)
        binheader[12] = int(float(self.header.get("OMEGA_START", 1)) * 1e3)
        binheader[13] = int(float(self.header.get("OMEGA_END", 1)) * 1e3)
        binheader[14] = int(float(self.header.get("CHI", 1)) * 1e3)
        binheader[15] = int(float(self.header.get("TWOTHETA", 1)) * 1e3)
        self.header["HIGH"] = str(binheader[2])
        if self.swap_needed:
            binheader.byteswap(True)
        return binheader.tobytes()

    def ascii_header(self, linesep="\n", size=4096):
        """
        Generate the ASCII header for writing

        :param linesep: end of line separator
        :param size: size of the header (without the binary header)
        :return: string (unicode) containing the mar345 header

        """
        version = fabio.version
        lnsep = len(linesep)

        dim2, dim1 = self.shape
        lstout = ['mar research'.ljust(64 - lnsep)]
        lstout.append("PROGRAM".ljust(15) + (str(self.header.get("PROGRAM", "FabIO Version %s" % (version))).ljust(49 - lnsep)))
        lstout.append("DATE".ljust(15) + (str(self.header.get("DATE", time.ctime()))).ljust(49 - lnsep))
        key = "SCANNER"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key = "FORMAT_TYPE"
        if key in self.header:
            lstout.append("FORMAT".ljust(15) + ("%s  %s %s" % (dim1, self.header[key], dim1 * dim2)).ljust(49 - lnsep))
        key = "HIGH"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key1 = "PIXEL_LENGTH"
        key2 = "PIXEL_HEIGHT"
        if (key1 in self.header) and (key2 in self.header):
            lstout.append("PIXEL".ljust(15) + ("LENGTH %s  HEIGHT %s" % (self.header[key1], self.header[key2])).ljust(49 - lnsep))
        key1 = "OFFSET_ROFF"
        key2 = "OFFSET_TOFF"
        if key1 in self.header and key2 in self.header:
            lstout.append("OFFSET".ljust(15) + ("ROFF %s  TOFF %s" % (self.header[key1], self.header[key2])).ljust(49 - lnsep))
        key = "MULTIPLIER"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key = "GAIN"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key = "WAVELENGTH"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key = "DISTANCE"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key = "RESOLUTION"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key1 = "PHI_START"
        key2 = "PHI_END"
        key3 = "PHI_OSC"
        if (key1 in self.header) and (key2 in self.header) and (key3 in self.header):
            lstout.append("PHI".ljust(15) + ("START %s  END %s  OSC %s" % (self.header[key1], self.header[key2], self.header[key3])).ljust(49 - lnsep))
        key1 = "OMEGA_START"
        key2 = "OMEGA_END"
        key3 = "OMEGA_OSC"
        if (key1 in self.header) and (key2 in self.header) and (key3 in self.header):
            lstout.append("OMEGA".ljust(15) + ("START %s  END %s  OSC %s" % (self.header[key1], self.header[key2], self.header[key3])).ljust(49 - lnsep))
        key = "CHI"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key = "TWOTHETA"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key1 = "CENTER_X"
        key2 = "CENTER_Y"
        if (key1 in self.header) and (key2 in self.header):
            lstout.append("CENTER".ljust(15) + ("X %s  Y %s" % (self.header[key1], self.header[key2])).ljust(49 - lnsep))
        key = "MODE"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key = "TIME"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key1 = "COUNTS_START"
        key2 = "COUNTS_END"
        key3 = "COUNTS_NMEAS"
        if key1 in self.header and key2 in self.header and key3 in self.header:
            lstout.append("COUNTS".ljust(15) + ("START %s  END %s  NMEAS %s" % (self.header[key1], self.header[key2], self.header[key3])).ljust(49 - lnsep))
        key1 = "COUNTS_MIN"
        key2 = "COUNTS_MAX"
        if key1 in self.header and key2 in self.header:
            lstout.append("COUNTS".ljust(15) + ("MIN %s  MAX %s" % (self.header[key1], self.header[key2])).ljust(49 - lnsep))
        key1 = "COUNTS_AVE"
        key2 = "COUNTS_SIG"
        if key1 in self.header and key2 in self.header:
            lstout.append("COUNTS".ljust(15) + ("AVE %s  SIG %s" % (self.header[key1], self.header[key2])).ljust(49 - lnsep))
        key1 = "INTENSITY_MIN"
        key2 = "INTENSITY_MAX"
        key3 = "INTENSITY_AVE"
        key4 = "INTENSITY_SIG"
        if key1 in self.header and key2 in self.header and key3 in self.header and key4 in self.header:
            lstout.append("INTENSITY".ljust(15) + ("MIN %s  MAX %s  AVE %s  SIG %s" % (self.header[key1], self.header[key2], self.header[key3], self.header[key4])).ljust(49 - lnsep))
        key1 = "HISTOGRAM_START"
        key2 = "HISTOGRAM_END"
        key3 = "HISTOGRAM_MAX"
        if key1 in self.header and key2 in self.header and key3 in self.header:
            lstout.append("HISTOGRAM".ljust(15) + ("START %s  END %s  MAX %s" % (self.header[key1], self.header[key2], self.header[key3])).ljust(49 - lnsep))
        key = "GENERATOR"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key = "MONOCHROMATOR"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        key1 = "COLLIMATOR_WIDTH"
        key2 = "COLLIMATOR_HEIGHT"
        if key1 in self.header and key2 in self.header:
            lstout.append("COLLIMATOR".ljust(15) + ("WIDTH %s  HEIGHT %s" % (self.header[key1], self.header[key2])).ljust(49 - lnsep))
        key = "REMARK"
        if key in self.header:
            lstout.append(key.ljust(15) + str(self.header[key]).ljust(49 - lnsep))
        else:
            lstout.append(key.ljust(64 - lnsep))
        key = "END OF HEADER"
        lstout.append(key)
        return linesep.join(lstout).ljust(size)

    def _high_intensity_pixel_records(self):
        flt_data = self.data.flatten()
        pix_location = numpy.where(flt_data > 65535)[0]
        nb_pix = pix_location.size
        if nb_pix % 8 == 0:
            tmp = numpy.zeros((nb_pix, 2), dtype="int32")
        else:
            tmp = numpy.zeros(((nb_pix // 8 + 1) * 8, 2), dtype="int32")
        tmp[:nb_pix, 0] = pix_location + 1
        tmp[:nb_pix, 1] = flt_data[pix_location]
        if self.swap_needed:
            tmp.byteswap(True)
        return tmp.tobytes()

    def nb_overflow_pixels(self):
        return (self.data > 65535).sum()

    @staticmethod
    def checkData(data=None):
        if data is None:
            return None
        else:
            #            enforce square image
            shape = data.shape
            assert len(shape) == 2, "image has 2 dimensions"
            mshape = max(shape)
            z = numpy.zeros((mshape, mshape), dtype=int)
            z[:shape[0], :shape[1]] = data
            return z


mar345image = Mar345Image



#!/usr/bin/env python
# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation files
#  (the "Software"), to deal in the Software without restriction,
#  including without limitation the rights to use, copy, modify, merge,
#  publish, distribute, sublicense, and/or sell copies of the Software,
#  and to permit persons to whom the Software is furnished to do so,
#  subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
#  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
#  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#  OTHER DEALINGS IN THE SOFTWARE.

"""Portable image converter based on FabIO library
to export Eiger frames (including te one from LIMA)
to CBF and mimic the header from Dectris Pilatus.
"""

__author__ = "Jerome Kieffer"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__licence__ = "MIT"
__date__ = "23/04/2021"
__status__ = "production"

import logging
logging.basicConfig()
logger = logging.getLogger("eiger2cbf")
import codecs
import sys
import os

from ..utils.cli import expand_args
from ..openimage import openimage as fabio_open
from .. import cbfimage, limaimage, eigerimage, version as fabio_version
import numpy
import argparse
try:
    import hdf5plugin
except ImportError:
    pass

try:
    import numexpr
except:
    logger.error("Numexpr is needed to interpret formula ...")

logger = logging.getLogger("eiger2cbf")
EXIT_SUCCESS = 0
EXIT_FAILURE = 1
EXIT_ARGUMENT_FAILURE = 2

try:
    from scipy import constants
except ImportError:
    # Updated with scipy 1.4
    CONST_hc = 12.398419843320026
else:
    CONST_hc = constants.c * constants.h / constants.e * 1e7


class ProgressBar:
    """
    Progress bar in shell mode
    """

    def __init__(self, title, max_value, bar_width):
        """
        Create a progress bar using a title, a maximum value and a graphical size.

        The display is done with stdout using carriage return to to hide the
        previous progress. It is not possible to use stdout for something else
        whill a progress bar is in use.

        The result looks like:

        .. code-block:: none

            Title [■■■■■■      ]  50%  Message

        :param str title: Title displayed before the progress bar
        :param float max_value: The maximum value of the progress bar
        :param int bar_width: Size of the progressbar in the screen
        """
        self.title = title
        self.max_value = max_value
        self.bar_width = bar_width
        self.last_size = 0
        self._message = ""
        self._value = 0.0

        encoding = None
        if hasattr(sys.stdout, "encoding"):
            # sys.stdout.encoding can't be used in unittest context with some
            # configurations of TestRunner. It does not exists in Python2
            # StringIO and is None in Python3 StringIO.
            encoding = sys.stdout.encoding
        if encoding is None:
            # We uses the safer aproch: a valid ASCII character.
            self.progress_char = '#'
        else:
            try:
                import datetime
                if str(datetime.datetime.now())[5:10] == "02-14":
                    self.progress_char = u'\u2665'
                else:
                    self.progress_char = u'\u25A0'
                _byte = codecs.encode(self.progress_char, encoding)
            except (ValueError, TypeError, LookupError):
                # In case the char is not supported by the encoding,
                # or if the encoding does not exists
                self.progress_char = '#'

    def clear(self):
        """
        Remove the progress bar from the display and move the cursor
        at the beginning of the line using carriage return.
        """
        sys.stdout.write('\r' + " " * self.last_size + "\r")
        sys.stdout.flush()

    def display(self):
        """
        Display the progress bar to stdout
        """
        self.update(self._value, self._message)

    def update(self, value, message="", max_value=None):
        """
        Update the progrss bar with the progress bar's current value.

        Set the progress bar's current value, compute the percentage
        of progress and update the screen with. Carriage return is used
        first and then the content of the progress bar. The cursor is
        at the begining of the line.

        :param float value: progress bar's current value
        :param str message: message displayed after the progress bar
        :param float max_value: If not none, update the maximum value of the
            progress bar
        """
        if max_value is not None:
            self.max_value = max_value
        self._message = message
        self._value = value

        if self.max_value == 0:
            coef = 1.0
        else:
            coef = (1.0 * value) / self.max_value
        percent = round(coef * 100)
        bar_position = int(coef * self.bar_width)
        if bar_position > self.bar_width:
            bar_position = self.bar_width

        # line to display
        line = '\r%15s [%s%s] % 3d%%  %s' % (self.title, self.progress_char * bar_position, ' ' * (self.bar_width - bar_position), percent, message)

        # trailing to mask the previous message
        line_size = len(line)
        clean_size = self.last_size - line_size
        if clean_size < 0:
            clean_size = 0
        self.last_size = line_size

        sys.stdout.write(line + " " * clean_size + "\r")
        sys.stdout.flush()


def select_detecor(shape):
    """CrysalisPro only accepts some detector shapes as CBF inputs.
    Those shaped correspond to the one of the dectris detector Pilatus and Eiger (first generation only?)
    
    This function takes the input shape and return the smallest Dectris shape which is larger ... 
    """
    assert len(shape) >= 2
    valid_detectors = {  # (514, 1030): 'eiger_500k',
                         # (1065, 1030): 'eiger_1m',
                         # (2167, 2070): 'eiger_4m',
                         # (3269, 3110): 'eiger_9m',
                         # (4371, 4150): 'eiger_16m',
                         # (512, 1028): 'eiger2cdte500k',
                         # (1062, 1028): 'eiger2cdte1m',
                         # (2162, 2068): 'eiger2cdte4m',
                         # (3262, 3108): 'eiger2cdte9m',
                         # (4362, 4148): 'eiger2cdte16m',
                         # (1, 1280): 'mythen1280',
                         (195, 487): 'pilatus_100k',
                         (407, 487): 'pilatus_200k',
                         (619, 487): 'pilatus_300k',
                         (195, 1475): 'pilatus_300kw',
                         (1043, 981): 'pilatus_1m',
                         (1679, 1475): 'pilatus_2m',
                         (2527, 2463): 'pilatus_6m',
                         # (195, 4439): 'pilatus_900kw'
                         }
    best = None
    for k in valid_detectors:
        if k[0] >= shape[-2] and k[1] >= shape[-1]:
            if best is None:
                best = k
            elif k[0] < best[0] or k[1] < best[1]:
                best = k
    return best


def convert_one(input_filename, options, start_at=0):
    """
    Convert a single file using options

    :param str input_filename: The input filename
    :param object options: List of options provided from the command line
    :param start_at: index to start at for given file
    :rtype: int
    :returns: the number of frames processed
    """
    flip = bool((options.rotation // 90) % 2)
    if options.transpose:
        flip = not flip

    input_filename = os.path.abspath(input_filename)
    input_exists = os.path.exists(input_filename)

    if options.verbose:
        print("Converting file '%s'" % (input_filename))

    if not input_exists:
        logger.error("Input file '%s' do not exists. Conversion skipped.", input_filename)
        return -1

    try:
        logger.debug("Load '%s'", input_filename)
        source = fabio_open(input_filename)
    except KeyboardInterrupt:
        raise
    except Exception as e:
        logger.error("Loading input file '%s' failed cause: \"%s\". Conversion skipped.", input_filename, e.message)
        logger.debug("Backtrace", exc_info=True)
        return -1

    shape = select_detecor((source.shape[-1], source.shape[-2]) if flip else source.shape)
    pilatus_headers = cbfimage.PilatusHeader("Silicon sensor, thickness 0.001 m")
    if isinstance(source, limaimage.LimaImage):
        # Populate the Pilatus header from the Lima
        entry_name = source.h5.attrs.get("default")
        if entry_name:
            entry = source.h5.get(entry_name)
            if entry:
                data_name = entry.attrs["default"]
                if data_name:
                    data_grp = entry.get(data_name)
                    if data_grp:
                        nxdetector = data_grp.parent
                        try:
                            detector = "%s, S/N %s" % (nxdetector["detector_information/model"][()],
                                                 nxdetector["detector_information/name"][()])
                            pilatus_headers["Detector"] = detector
                        except Exception as e:
                            logger.warning("Error in searching for detector definition (%s): %s", type(e), e)
                        try:
                            pilatus_headers["Pixel_size"] = (nxdetector["detector_information/pixel_size/xsize"][()],
                                                             nxdetector["detector_information/pixel_size/ysize"][()])
                        except Exception as e:
                            logger.warning("Error in searching for pixel size (%s): %s", type(e), e)
                        try:
                            t1 = nxdetector["acquisition/exposure_time"][()]
                            t2 = nxdetector["acquisition/latency_time"][()]
                            pilatus_headers["Exposure_time"] = t1
                            pilatus_headers["Exposure_period"] = t1 + t2
                        except Exception as e:
                            logger.warning("Error in searching for exposure time (%s): %s", type(e), e)
    elif isinstance(source, eigerimage.EigerImage):
        raise NotImplementedError("Please implement Eiger detector data format parsing or at least open an issue")
    else:
        raise NotImplementedError("Unsupported format: %s" % source.__class__.__name__)

    # Parse option for Pilatus headers
    if options.energy:
        pilatus_headers["Wavelength"] = CONST_hc / options.energy
    elif options.wavelength:
        pilatus_headers["Wavelength"] = options.wavelength
    if options.distance:
        pilatus_headers["Detector_distance"] = options.distance
    if options.beam:
        pilatus_headers["Beam_xy"] = options.beam
    if options.alpha:
        pilatus_headers["Alpha"] = options.alpha
    if options.kappa:
        pilatus_headers["Kappa"] = options.kappa
    formula = None
    destination = None
    if options.chi is not None:
        try:
            value = float(options.chi)
        except ValueError:
            # Handle the string
            formula = numexpr.NumExpr(options.chi)
            destination = "Chi"
            pilatus_headers["Oscillation_axis"] = "CHI"
        else:
            pilatus_headers["Chi"] = value
            pilatus_headers["Chi_increment"] = 0.0

    if options.phi is not None:
        try:
            value = float(options.phi)
        except ValueError:
            # Handle the string
            formula = numexpr.NumExpr(options.phi)
            destination = "Phi"
            pilatus_headers["Oscillation_axis"] = "PHI"
        else:
            pilatus_headers["Phi"] = value
            pilatus_headers["Phi_increment"] = 0.0
    if options.omega is not None:
        try:
            value = float(options.omega)
        except ValueError:
            # Handle the string
            formula = numexpr.NumExpr(options.omega)
            destination = "Omega"
            pilatus_headers["Oscillation_axis"] = "OMEGA"
        else:
            pilatus_headers["Omega"] = value
            pilatus_headers["Omega_increment"] = 0.0

    for i, frame in enumerate(source):
        idx = i + start_at
        data = numpy.empty(shape, dtype=numpy.int32)
        data.fill(options.dummy)
        input_data = frame.data.astype(numpy.int32)
        if options.rotation:
            input_data = numpy.rot90(input_data, k=options.rotation // 90)
        if options.transpose:
            input_data = input_data.T
        if options.flip_ud:
            input_data = numpy.flipud(input_data)
        if options.flip_lr:
            input_data = numpy.fliplr(input_data)

        data[:input_data.shape[0],:input_data.shape[1]] = input_data

        mask = numpy.where(input_data == numpy.iinfo(frame.data.dtype).max)
        data[mask] = options.dummy
        converted = cbfimage.CbfImage(data=data)

        if formula and destination:
            position = formula(idx)
            delta = (formula(idx + 1) - position)
            pilatus_headers["Start_angle"] = pilatus_headers[destination] = position
            pilatus_headers["Angle_increment"] = pilatus_headers[destination + "_increment"] = delta
        converted.pilatus_headers = pilatus_headers

        output_filename = options.output.format(index=((idx + options.offset)))
        os.makedirs(os.path.dirname(output_filename), exist_ok=True)
        try:
            logger.debug("Write '%s'", output_filename)
            if not options.dry_run:
                converted.write(output_filename)
        except KeyboardInterrupt:
            raise
        except Exception as e:
            logger.error("Saving output file '%s' failed cause: \"%s: %s\". Conversion skipped.", output_filename, type(e), e)
            logger.debug("Backtrace", exc_info=True)
            return -1
    return source.nframes


def convert_all(options):
    """Convert all the files from the command line.

    :param object options: List of options provided from the command line
    :rtype: bool
    :returns: True is the conversion succeeded
    """
    pb = ProgressBar("HDF5 --> CBFs", len(options.images), 30)
    succeeded = True
    start_at = 0
    for i, filename in enumerate(options.images):
        pb.update(i + 0.5, os.path.basename(filename))
        finish_at = convert_one(filename, options, start_at)
        succeeded = succeeded and (finish_at > 0)
        start_at += finish_at
    pb.clear()
    return succeeded


def main():

    epilog = """return codes: 0 means a success. 1 means the conversion
                contains a failure, 2 means there was an error in the
                arguments"""

    parser = argparse.ArgumentParser(prog="eiger2cbf",
                                     description=__doc__,
                                     epilog=epilog)
    parser.add_argument("IMAGE", nargs="*",
                        help="File with input images")
    parser.add_argument("-V", "--version", action='version', version=fabio_version,
                        help="output version and exit")
    parser.add_argument("-v", "--verbose", action='store_true', dest="verbose", default=False,
                        help="show information for each conversions")
    parser.add_argument("--debug", action='store_true', dest="debug", default=False,
                        help="show debug information")
    group = parser.add_argument_group("main arguments")
#     group.add_argument("-l", "--list", action="store_true", dest="list", default=None,
#                        help="show the list of available formats and exit")
    group.add_argument("-o", "--output", default='eiger2cbf/frame_{index:04d}.cbf', type=str,
                       help="output directory and filename template")
    group.add_argument("-O", "--offset", type=int, default=0,
                       help="index offset, CrysalisPro likes indexes to start at 1, Python starts at 0")
    group.add_argument("-D", "--dummy", type=int, default=-1,
                       help="Set masked values to this dummy value")

    group = parser.add_argument_group("optional behaviour arguments")
#     group.add_argument("-f", "--force", dest="force", action="store_true", default=False,
#                        help="if an existing destination file cannot be" +
#                        " opened, remove it and try again (this option" +
#                        " is ignored when the -n option is also used)")
#     group.add_argument("-n", "--no-clobber", dest="no_clobber", action="store_true", default=False,
#                        help="do not overwrite an existing file (this option" +
#                        " is ignored when the -i option is also used)")
#     group.add_argument("--remove-destination", dest="remove_destination", action="store_true", default=False,
#                        help="remove each existing destination file before" +
#                        " attempting to open it (contrast with --force)")
#     group.add_argument("-u", "--update", dest="update", action="store_true", default=False,
#                        help="copy only when the SOURCE file is newer" +
#                        " than the destination file or when the" +
#                        " destination file is missing")
#     group.add_argument("-i", "--interactive", dest="interactive", action="store_true", default=False,
#                        help="prompt before overwrite (overrides a previous -n" +
#                        " option)")
    group.add_argument("--dry-run", dest="dry_run", action="store_true", default=False,
                       help="do everything except modifying the file system")

    group = parser.add_argument_group("Experimental setup options")
    group.add_argument("-e", "--energy", type=float, default=None,
                       help="Energy of the incident beam in keV")
    group.add_argument("-w", "--wavelength", type=float, default=None,
                       help="Wavelength of the incident beam in Å")
    group.add_argument("-d", "--distance", type=float, default=None,
                       help="Detector distance in meters")
    group.add_argument("-b", "--beam", nargs=2, type=float, default=None,
                       help="Direct beam in pixels x, y")

    group = parser.add_argument_group("Goniometer setup")
#     group.add_argument("--axis", type=str, default=None,
#                        help="Goniometer angle used for scanning: 'omega', 'phi' or 'chi'")
    group.add_argument("--alpha", type=float, default=None,
                       help="Goniometer angle alpha value in deg.")
    group.add_argument("--kappa", type=float, default=None,
                       help="Goniometer angle kappa value in deg.")
    group.add_argument("--chi", type=str, default=None,
                       help="Goniometer angle chi value in deg. or formula f(index)")
    group.add_argument("--phi", type=str, default=None,
                       help="Goniometer angle phi value in deg. or formula f(index)")
    group.add_argument("--omega", type=str, default=None,
                       help="Goniometer angle omega value in deg. or formula f(index)")

    group = parser.add_argument_group("Image preprocessing (Important: applied in this order!)")
    group.add_argument("--rotation", type=int, default=0,
                       help="Rotate the initial image by this value in degrees. Must be a multiple of 90°.")
    group.add_argument("--transpose", default=False, action="store_true",
                       help="Flip the x/y axis")
    group.add_argument("--flip-ud", dest="flip_ud", default=False, action="store_true",
                       help="Flip the image upside-down")
    group.add_argument("--flip-lr", dest="flip_lr", default=False, action="store_true",
                       help="Flip the image left-right")

    try:
        args = parser.parse_args()

        if args.debug:
            logger.setLevel(logging.DEBUG)

#         if args.list:
#             print_supported_formats()
#             return

        if len(args.IMAGE) == 0:
            raise argparse.ArgumentError(None, "No input file specified.")

        # the upper case IMAGE is used for the --help auto-documentation
        args.images = expand_args(args.IMAGE)
        args.images.sort()
    except argparse.ArgumentError as e:
        logger.error(e.message)
        logger.debug("Backtrace", exc_info=True)
        return EXIT_ARGUMENT_FAILURE

    succeeded = convert_all(args)
    if not succeeded:
        print("Conversion or part of it failed. You can try with --debug to have more output information.")
        return EXIT_FAILURE

    return EXIT_SUCCESS


if __name__ == "__main__":
    result = main()
    sys.exit(result)



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "03/04/2020"

import re
import logging
from . import tifimage

_logger = logging.getLogger(__name__)


class PilatusTiffFrame(tifimage.TiffFrame):
    """Frame container for TIFF format generated by a Pilatus detector"""

    def __init__(self, data, tiff_header, pilatus_header):
        super(PilatusTiffFrame, self).__init__(data, tiff_header)
        self.pilatus_header = pilatus_header
        # Override the header
        self._header = pilatus_header


class PilatusImage(tifimage.TifImage):
    """ Read in Pilatus format, also
        pilatus images, including header info """

    DESCRIPTION = "Pilatus file format based on Tiff"

    DEFAULT_EXTENSIONS = ["tif", "tiff"]

    _keyvalue_spliter = re.compile(r"\s*[,:=\s]\s*")
    """It allow to split the first white space, colon, coma, or equal
    character and remove white spaces around"""

    def _create_pilatus_header(self, tiff_header):
        """
        Parse Pilatus header from a TIFF header.

        The Pilatus header is stored in the metadata ImageDescription (tag 270)
        as an ASCII text which looks like:

        .. block-code:: python

            imageDescription = '# Pixel_size 172e-6 m x 172e-6 m\r\n'\
                '# Silicon sensor, thickness 0.000320 m\r\n# Exposure_time 90.000000 s\r\n'\
                '# Exposure_period 90.000000 s\r\n# Tau = 0 s\r\n'\
                '# Count_cutoff 1048574 counts\r\n# Threshold_setting 0 eV\r\n'\
                '# Gain_setting not implemented (vrf = 9.900)\r\n'\
                '# N_excluded_pixels = 0\r\n# Excluded_pixels: (nil)\r\n'\
                '# Flat_field: (nil)\r\n# Trim_directory: (nil)\r\n\x00'

        :rtype: OrderedDict
        """
        if "imageDescription" not in tiff_header:
            # It is not a Pilatus TIFF image
            raise IOError("Image is not a Pilatus image")

        header = self.check_header()

        description = tiff_header["imageDescription"]
        for line in description.split("\n"):
            index = line.find('# ')
            if index == -1:
                if line.strip(" \x00") != "":
                    # If it is not an empty line
                    _logger.debug("Pilatus header line '%s' misformed. Skipped", line)
                continue

            line = line[2:].strip()
            if line == "":
                # empty line
                continue

            result = self._keyvalue_spliter.split(line, 1)
            if len(result) != 2:
                _logger.debug("Pilatus header line '%s' misformed. Skipped", line)
                continue

            key, value = result
            header[key] = value

        return header

    def _create_frame(self, image_data, tiff_header):
        """Create exposed data from TIFF information"""
        pilatus_header = self._create_pilatus_header(tiff_header)
        frame = PilatusTiffFrame(image_data, tiff_header, pilatus_header)
        return frame


pilatusimage = PilatusImage



# coding: utf-8
#
#    Project: FabIO X-ray image reader
#
#    Copyright (C) 2010-2016 European Synchrotron Radiation Facility
#                       Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""
FabIO class for dealing with JPEG 2000 images.
"""

__authors__ = ["Valentin Valls"]
__date__ = "06/04/2020"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__status__ = "stable"

import logging
logger = logging.getLogger(__name__)

try:
    import PIL
    import PIL.Image
except ImportError:
    PIL = None

try:
    import glymur
except ImportError:
    glymur = None

from .fabioimage import FabioImage
from .fabioutils import OrderedDict
from .utils import pilutils


class Jpeg2KImage(FabioImage):
    """
    Images in JPEG 2000 format.

    It uses PIL or glymur libraries.
    """
    DESCRIPTION = "JPEG 2000 format"

    DEFAULT_EXTENSIONS = ["jp2", "jpx", "j2k", "jpf", "jpg2"]

    _need_a_seek_to_read = True

    def __init__(self, *args, **kwds):
        """ Tifimage constructor adds an nbits member attribute """
        self.nbits = None
        FabioImage.__init__(self, *args, **kwds)
        self.lib = ""

        self._decoders = OrderedDict()
        if PIL is not None:
            self._decoders["PIL"] = self._readWithPil
        if glymur is not None:
            self._decoders["glymur"] = self._readWithGlymur

    def _readWithPil(self, filename, infile):
        """Read data using PIL"""
        pilimage = PIL.Image.open(infile)
        data = pilutils.get_numpy_array(pilimage)
        self.data = data

        if pilimage and pilimage.info:
            for k, v in pilimage.info.items():
                self.header[k] = v

    def _loadGlymurImage(self, filename, infile):
        """
        Hack to use Glymur with Python file object

        This code was tested with all release 0.8.x
        """
        # image = glymur.Jp2k(filename)
        version = tuple(int(i) for i in glymur.__version__.split(".")[:2])
        if version == (0, 7):
            image = glymur.Jp2k(filename=filename)
        elif version >= (0,8):
            # inject a shape  to avoid calling the read function
            image = glymur.Jp2k(filename=filename, shape=(1, 1))
        else:
            raise IOError("Glymur version %s is not supported" % glymur.__version__)

        # Move to the end of the file to know the size
        infile.seek(0, 2)
        length = infile.tell()
        infile.seek(0)

        # initialize what it should already be done
        image.length = length
        image._shape = None
        # It is not the only one format supported by Glymur
        # but it is a simplification
        image._codec_format = glymur.lib.openjp2.CODEC_JP2

        # parse the data
        image.box = image.parse_superbox(infile)
        try:
            image._validate()
        except Exception:
            logger.debug("Backtrace", exc_info=True)
            raise IOError("File %s is not a valid format" % filename)

        # Now the image can be used normaly
        return image

    def _readWithGlymur(self, filename, infile):
        """Read data using Glymur"""
        image = self._loadGlymurImage(filename, infile)
        self.data = image.read()

    def read(self, filename, frame=None):
        infile = self._open(filename, "rb")
        self.data = None

        for name, read in self._decoders.items():
            try:
                infile.seek(0)
                read(filename, infile)
                self.lib = name
                break

            except IOError as e:
                self.data = None
                self.header = OrderedDict()
                logger.debug("Error while using %s library: %s" % (name, e), exc_info=True)

        if self.data is None:
            infile.seek(0)
            raise IOError("No decoder available for the file %s." % filename)
        self.resetvals()
        return self


jpeg2kimage = Jpeg2KImage



# The PyMca X-Ray Fluorescence Toolkit
#
# Copyright (c) 2004-2020 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "V.A. Sole - ESRF Data Analysis"
__contact__ = "sole@esrf.fr"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "26/12/2020"

import sys
import os
import struct
import numpy
import logging

ALLOW_MULTIPLE_STRIPS = False

TAG_ID = {256: "NumberOfColumns",  # S or L ImageWidth
          257: "NumberOfRows",  # S or L ImageHeight
          258: "BitsPerSample",  # S Number of bits per component
          259: "Compression",  # SHORT (1 - NoCompression, ...
          262: "PhotometricInterpretation",  # SHORT (0 - WhiteIsZero, 1 -BlackIsZero, 2 - RGB, 3 - Palette color
          270: "ImageDescription",  # ASCII
          272: "Model",  # ASCII
          273: "StripOffsets",  # S or L, for each strip, the byte offset of the strip
          277: "SamplesPerPixel",  # SHORT (>=3) only for RGB images
          278: "RowsPerStrip",  # S or L, number of rows in each back may be not for the last
          279: "StripByteCounts",  # S or L, The number of bytes in the strip AFTER any compression
          305: "Software",  # ASCII
          306: "Date",  # ASCII
          320: "Colormap",  # Colormap of Palette-color Images
          339: "SampleFormat",  # SHORT Interpretation of data in each pixel
          }

# TILES ARE TO BE SUPPORTED TOO ...
TAG_NUMBER_OF_COLUMNS = 256
TAG_NUMBER_OF_ROWS = 257
TAG_BITS_PER_SAMPLE = 258
TAG_PHOTOMETRIC_INTERPRETATION = 262
TAG_COMPRESSION = 259
TAG_IMAGE_DESCRIPTION = 270
TAG_MODEL = 272
TAG_STRIP_OFFSETS = 273
TAG_SAMPLES_PER_PIXEL = 277
TAG_ROWS_PER_STRIP = 278
TAG_STRIP_BYTE_COUNTS = 279
TAG_SOFTWARE = 305
TAG_DATE = 306
TAG_COLORMAP = 320
TAG_SAMPLE_FORMAT = 339

FIELD_TYPE = {1: ('BYTE', "B"),
              2: ('ASCII', "s"),  # string ending with binary zero
              3: ('SHORT', "H"),
              4: ('LONG', "I"),
              5: ('RATIONAL', "II"),
              6: ('SBYTE', "b"),
              7: ('UNDEFINED', "B"),
              8: ('SSHORT', "h"),
              9: ('SLONG', "i"),
              10: ('SRATIONAL', "ii"),
              11: ('FLOAT', "f"),
              12: ('DOUBLE', "d")}

FIELD_TYPE_OUT = {'B': 1,
                  's': 2,
                  'H': 3,
                  'I': 4,
                  'II': 5,
                  'b': 6,
                  'h': 8,
                  'i': 9,
                  'ii': 10,
                  'f': 11,
                  'd': 12}

# sample formats (http://www.awaresystems.be/imaging/tiff/tiffflags/sampleformat.html)
SAMPLE_FORMAT_UINT = 1
SAMPLE_FORMAT_INT = 2
SAMPLE_FORMAT_FLOAT = 3  # floating point
SAMPLE_FORMAT_VOID = 4  # undefined data, usually assumed UINT
SAMPLE_FORMAT_COMPLEXINT = 5
SAMPLE_FORMAT_COMPLEXIEEEFP = 6

logger = logging.getLogger(__name__)


class TiffIO(object):

    def __init__(self, filename, mode=None, cache_length=20, mono_output=False):
        if mode is None:
            mode = 'rb'
        if 'b' not in mode:
            mode = mode + 'b'
        if 'a' in mode.lower():
            raise IOError("Mode %s makes no sense on TIFF files. Consider 'rb+'" % mode)
        if ('w' in mode):
            if '+' not in mode:
                mode += '+'

        if hasattr(filename, "seek") and\
           hasattr(filename, "read"):
            fd = filename
            self._access = None
        else:
            # the b is needed for windows and python 3
            fd = open(filename, mode)
            self._access = mode

        self._initInternalVariables(fd)
        self._maxImageCacheLength = cache_length
        self._forceMonoOutput = mono_output

    def __enter__(self):
        return self

    def __exit__(self, *arg):
        # TODO: inspace type, value and traceback
        self.close()

    def _initInternalVariables(self, fd=None):
        if fd is None:
            fd = self.fd
        else:
            self.fd = fd
        # read the order
        fd.seek(0)
        order = fd.read(2).decode()
        if len(order):
            if order == "II":
                # intel, little endian
                fileOrder = "little"
                self._structChar = '<'
            elif order == "MM":
                # motorola, high endian
                fileOrder = "big"
                self._structChar = '>'
            else:
                raise IOError("File is not a Mar CCD file, nor a TIFF file")
            a = fd.read(2)
            fortyTwo = struct.unpack(self._structChar + "H", a)[0]
            if fortyTwo != 42:
                raise IOError("Invalid TIFF version %d" % fortyTwo)
            else:
                logger.debug("VALID TIFF VERSION")
            if sys.byteorder != fileOrder:
                swap = True
            else:
                swap = False
        else:
            if sys.byteorder == "little":
                self._structChar = '<'
            else:
                self._structChar = '>'
            swap = False
        self._swap = swap
        self._IFD = []
        self._imageDataCacheIndex = []
        self._imageDataCache = []
        self._imageInfoCacheIndex = []
        self._imageInfoCache = []
        self.getImageFileDirectories(fd)

    def __makeSureFileIsOpen(self):
        if not self.fd.closed:
            return
        logger.debug("Reopening closed file")
        fileName = self.fd.name
        if self._access is None:
            # we do not own the file
            # open in read mode
            newFile = open(fileName, 'rb')
        else:
            newFile = open(fileName, self._access)
        self.fd = newFile

    def __makeSureFileIsClosed(self):
        if self._access is None:
            # we do not own the file
            logger.debug("Not closing not owned file")
            return

        if not self.fd.closed:
            self.fd.close()

    def close(self):
        return self.__makeSureFileIsClosed()

    def getNumberOfImages(self):
        # update for the case someone has done anything?
        self._updateIFD()
        return len(self._IFD)

    def _updateIFD(self):
        self.__makeSureFileIsOpen()
        self.getImageFileDirectories()
        self.__makeSureFileIsClosed()

    def getImageFileDirectories(self, fd=None):
        if fd is None:
            fd = self.fd
        else:
            self.fd = fd
        st = self._structChar
        fd.seek(4)
        self._IFD = []
        nImages = 0
        fmt = st + 'I'
        inStr = fd.read(struct.calcsize(fmt))
        if not len(inStr):
            offsetToIFD = 0
        else:
            offsetToIFD = struct.unpack(fmt, inStr)[0]
        logger.debug("Offset to first IFD = %d", offsetToIFD)
        while offsetToIFD != 0:
            self._IFD.append(offsetToIFD)
            nImages += 1
            fd.seek(offsetToIFD)
            fmt = st + 'H'
            numberOfDirectoryEntries = struct.unpack(fmt, fd.read(struct.calcsize(fmt)))[0]
            logger.debug("Number of directory entries = %d", numberOfDirectoryEntries)

            fmt = st + 'I'
            fd.seek(offsetToIFD + 2 + 12 * numberOfDirectoryEntries)
            offsetToIFD = struct.unpack(fmt, fd.read(struct.calcsize(fmt)))[0]
            logger.debug("Next Offset to IFD = %d", offsetToIFD)
            # offsetToIFD = 0
        logger.debug("Number of images found = %d", nImages)
        return nImages

    def _parseImageFileDirectory(self, nImage):
        offsetToIFD = self._IFD[nImage]
        st = self._structChar
        fd = self.fd
        fd.seek(offsetToIFD)
        fmt = st + 'H'
        numberOfDirectoryEntries = struct.unpack(fmt, fd.read(struct.calcsize(fmt)))[0]
        logger.debug("Number of directory entries = %d", numberOfDirectoryEntries)

        fmt = st + 'HHI4s'
        tagIDList = []
        fieldTypeList = []
        nValuesList = []
        valueOffsetList = []
        for _ in range(numberOfDirectoryEntries):
            tagID, fieldType, nValues, valueOffset = struct.unpack(fmt, fd.read(12))
            tagIDList.append(tagID)
            fieldTypeList.append(fieldType)
            nValuesList.append(nValues)
            if nValues == 1:
                ftype, vfmt = FIELD_TYPE[fieldType]
                if ftype not in ['ASCII', 'RATIONAL', 'SRATIONAL']:
                    vfmt = st + vfmt
                    data = valueOffset[0: struct.calcsize(vfmt)]
                    if struct.calcsize(vfmt) > len(data):
                        # Add a 0 padding to have the expected size
                        logger.warning("Data at tag id '%s' is smaller than expected", tagID)
                        data = data + b"\x00" * (struct.calcsize(vfmt) - len(data))
                    actualValue = struct.unpack(vfmt, data)[0]
                    valueOffsetList.append(actualValue)
                else:
                    valueOffsetList.append(valueOffset)
            elif (nValues < 5) and (fieldType == 2):
                ftype, vfmt = FIELD_TYPE[fieldType]
                vfmt = st + "%d%s" % (nValues, vfmt)
                actualValue = struct.unpack(vfmt, valueOffset[0: struct.calcsize(vfmt)])[0]
                valueOffsetList.append(actualValue)
            else:
                valueOffsetList.append(valueOffset)
            if logger.getEffectiveLevel() == logging.DEBUG:
                if tagID in TAG_ID:
                    logger.debug("tagID = %s", TAG_ID[tagID])
                else:
                    logger.debug("tagID        = %d", tagID)
                logger.debug("fieldType    = %s", FIELD_TYPE[fieldType][0])
                logger.debug("nValues      = %d", nValues)
                # if nValues == 1:
                #    logger.debug("valueOffset =  %s", valueOffset)
        return tagIDList, fieldTypeList, nValuesList, valueOffsetList

    def _readIFDEntry(self, tag, tagIDList, fieldTypeList, nValuesList, valueOffsetList):
        fd = self.fd
        st = self._structChar
        idx = tagIDList.index(tag)
        nValues = nValuesList[idx]
        output = []
        _ftype, vfmt = FIELD_TYPE[fieldTypeList[idx]]
        vfmt = st + "%d%s" % (nValues, vfmt)
        requestedBytes = struct.calcsize(vfmt)
        if nValues == 1:
            output.append(valueOffsetList[idx])
        elif requestedBytes < 5:
            output.append(valueOffsetList[idx])
        else:
            fd.seek(struct.unpack(st + "I", valueOffsetList[idx])[0])
            output = struct.unpack(vfmt, fd.read(requestedBytes))

        if fieldTypeList[idx] == 2:
            # That's an ASCII tag
            cleaned_output = []
            for raw in output:
                # remove the trailing \x00
                index = raw.find(b"\x00")
                if index != -1:
                    raw = raw[0:index]
                # read the data as text
                try:
                    text = raw.decode("utf-8")
                except UnicodeDecodeError:
                    logger.warning("TIFF file tag %d contains non ASCII/UTF-8 characters. ", tag)
                    text = raw.decode("utf-8", errors='replace')
                    # Use a valid ASCII character to limit ferther encoding error
                    text = text.replace(u"\ufffd", "?")

                cleaned_output.append(text)
            if isinstance(output, tuple):
                output = tuple(cleaned_output)
            else:
                output = cleaned_output
        return output

    def getData(self, nImage, **kw):
        if nImage >= len(self._IFD):
            # update prior to raise an index error error
            self._updateIFD()
        return self._readImage(nImage, **kw)

    def getImage(self, nImage):
        return self.getData(nImage)

    def getInfo(self, nImage, **kw):
        if nImage >= len(self._IFD):
            # update prior to raise an index error error
            self._updateIFD()
        # current = self._IFD[nImage]
        return self._readInfo(nImage)

    def _readInfo(self, nImage, close=True):
        if nImage in self._imageInfoCacheIndex:
            logger.debug("Reading info from cache")
            return self._imageInfoCache[self._imageInfoCacheIndex.index(nImage)]

        # read the header
        self.__makeSureFileIsOpen()
        tagIDList, fieldTypeList, nValuesList, valueOffsetList = self._parseImageFileDirectory(nImage)

        # rows and columns
        nColumns = valueOffsetList[tagIDList.index(TAG_NUMBER_OF_COLUMNS)]
        nRows = valueOffsetList[tagIDList.index(TAG_NUMBER_OF_ROWS)]

        # bits per sample
        idx = tagIDList.index(TAG_BITS_PER_SAMPLE)
        nBits = valueOffsetList[idx]
        if nValuesList[idx] != 1:
            # this happens with RGB and friends, nBits is not a single value
            nBits = self._readIFDEntry(TAG_BITS_PER_SAMPLE,
                                       tagIDList, fieldTypeList, nValuesList,
                                       valueOffsetList)

        if TAG_COLORMAP in tagIDList:
            idx = tagIDList.index(TAG_COLORMAP)
            tmpColormap = self._readIFDEntry(TAG_COLORMAP,
                                             tagIDList, fieldTypeList, nValuesList, valueOffsetList)
            if max(tmpColormap) > 255:
                tmpColormap = numpy.array(tmpColormap, dtype=numpy.uint16)
                tmpColormap = (tmpColormap / 256.).astype(numpy.uint8)
            else:
                tmpColormap = numpy.array(tmpColormap, dtype=numpy.uint8)
            tmpColormap.shape = 3, -1
            colormap = numpy.zeros((tmpColormap.shape[-1], 3), tmpColormap.dtype)
            colormap[:,:] = tmpColormap.T
            tmpColormap = None
        else:
            colormap = None

        # sample format
        if TAG_SAMPLE_FORMAT in tagIDList:
            sampleFormat = valueOffsetList[tagIDList.index(TAG_SAMPLE_FORMAT)]
        else:
            # set to unknown
            sampleFormat = SAMPLE_FORMAT_VOID

        # compression
        compression = False
        compression_type = 1
        if TAG_COMPRESSION in tagIDList:
            compression_type = valueOffsetList[tagIDList.index(TAG_COMPRESSION)]
            if compression_type == 1:
                compression = False
            else:
                compression = True

        # photometric interpretation
        interpretation = 1
        if TAG_PHOTOMETRIC_INTERPRETATION in tagIDList:
            interpretation = valueOffsetList[tagIDList.index(TAG_PHOTOMETRIC_INTERPRETATION)]
        else:
            logger.debug("WARNING: Non standard TIFF. Photometric interpretation TAG missing")
        helpString = ""

        if TAG_IMAGE_DESCRIPTION in tagIDList:
            imageDescription = self._readIFDEntry(TAG_IMAGE_DESCRIPTION,
                                                  tagIDList, fieldTypeList, nValuesList, valueOffsetList)
            if type(imageDescription) in [type([1]), type((1,))]:
                imageDescription = helpString.join(imageDescription)
        else:
            imageDescription = "%d/%d" % (nImage + 1, len(self._IFD))

        if TAG_MODEL in tagIDList:
            model = self._readIFDEntry(TAG_MODEL,
                                       tagIDList, fieldTypeList, nValuesList, valueOffsetList)
        else:
            model = None

        defaultSoftware = "Unknown Software"

        if TAG_SOFTWARE in tagIDList:
            software = self._readIFDEntry(TAG_SOFTWARE,
                                          tagIDList, fieldTypeList, nValuesList, valueOffsetList)
            if isinstance(software, (tuple, list)):
                software = helpString.join(software)
        else:
            software = defaultSoftware

        if software == defaultSoftware:
            try:
                if imageDescription.upper().startswith("IMAGEJ"):
                    software = imageDescription.split("=")[0]
            except Exception:
                pass

        if TAG_DATE in tagIDList:
            date = self._readIFDEntry(TAG_DATE,
                                      tagIDList, fieldTypeList, nValuesList, valueOffsetList)
            if type(date) in [type([1]), type((1,))]:
                date = helpString.join(date)
        else:
            date = "Unknown Date"

        stripOffsets = self._readIFDEntry(TAG_STRIP_OFFSETS,
                                          tagIDList,
                                          fieldTypeList,
                                          nValuesList,
                                          valueOffsetList)
        if TAG_ROWS_PER_STRIP in tagIDList:
            rowsPerStrip = self._readIFDEntry(TAG_ROWS_PER_STRIP,
                                              tagIDList,
                                              fieldTypeList,
                                              nValuesList,
                                              valueOffsetList)[0]
        else:
            rowsPerStrip = nRows
            logger.warning("Non standard TIFF. Rows per strip TAG missing")

        if TAG_STRIP_BYTE_COUNTS in tagIDList:
            stripByteCounts = self._readIFDEntry(TAG_STRIP_BYTE_COUNTS,
                                                 tagIDList,
                                                 fieldTypeList,
                                                 nValuesList,
                                                 valueOffsetList)
        else:
            logger.warning("Non standard TIFF. Strip byte counts TAG missing")
            if hasattr(nBits, 'index'):
                expectedSum = 0
                for n in nBits:
                    expectedSum += int(nRows * nColumns * n / 8)
            else:
                expectedSum = int(nRows * nColumns * nBits / 8)
            stripByteCounts = [expectedSum]

        if close:
            self.__makeSureFileIsClosed()

        if self._forceMonoOutput and (interpretation > 1):
            # color image but asked monochrome output
            nBits = 32
            colormap = None
            sampleFormat = SAMPLE_FORMAT_FLOAT
            interpretation = 1
            # we cannot rely on any cache in this case
            useInfoCache = False
            logger.debug("FORCED MONO")
        else:
            useInfoCache = True

        info = {}
        info["nRows"] = nRows
        info["nColumns"] = nColumns
        info["nBits"] = nBits
        info["compression"] = compression
        info["compression_type"] = compression_type
        info["imageDescription"] = imageDescription
        info["stripOffsets"] = stripOffsets  # This contains the file offsets to the data positions
        info["rowsPerStrip"] = rowsPerStrip
        info["stripByteCounts"] = stripByteCounts  # bytes in strip since I do not support compression
        info["software"] = software
        info["date"] = date
        info["colormap"] = colormap
        info["sampleFormat"] = sampleFormat
        info["photometricInterpretation"] = interpretation
        if model is not None:
            info["model"] = model

        infoDict = {}
        testString = 'PyMca'
        if software.startswith(testString):
            # str to make sure python 2.x sees it as string and not unicode
            descriptionString = imageDescription
            # interpret the image description in terms of supplied
            # information at writing time
            items = descriptionString.split('=')
            for i in range(int(len(items) / 2)):
                key = "%s" % items[i * 2]
                # get rid of the \n at the end of the value
                value = "%s" % items[i * 2 + 1][:-1]
                infoDict[key] = value
        info['info'] = infoDict

        if (self._maxImageCacheLength > 0) and useInfoCache:
            self._imageInfoCacheIndex.insert(0, nImage)
            self._imageInfoCache.insert(0, info)
            if len(self._imageInfoCacheIndex) > self._maxImageCacheLength:
                self._imageInfoCacheIndex = self._imageInfoCacheIndex[:self._maxImageCacheLength]
                self._imageInfoCache = self._imageInfoCache[:self._maxImageCacheLength]
        return info

    def _readImage(self, nImage, **kw):
        logger.debug("Reading image %d", nImage)
        if 'close' in kw:
            close = kw['close']
        else:
            close = True
        rowMin = kw.get('rowMin', None)
        rowMax = kw.get('rowMax', None)
        if nImage in self._imageDataCacheIndex:
            logger.debug("Reading image data from cache")
            return self._imageDataCache[self._imageDataCacheIndex.index(nImage)]

        self.__makeSureFileIsOpen()
        if self._forceMonoOutput:
            oldMono = True
        else:
            oldMono = False
        try:
            self._forceMonoOutput = False
            info = self._readInfo(nImage, close=False)
            self._forceMonoOutput = oldMono
        except Exception:
            logger.debug("Backtrace", exc_info=True)
            self._forceMonoOutput = oldMono
            raise
        compression = info['compression']
        compression_type = info['compression_type']
        if compression:
            if compression_type != 32773:
                raise IOError("Compressed TIFF images not supported except packbits")
            else:
                # PackBits compression
                logger.debug("Using PackBits compression")

        interpretation = info["photometricInterpretation"]
        if interpretation == 2:
            # RGB
            pass
            # raise IOError("RGB Image. Only grayscale images supported")
        elif interpretation == 3:
            # Palette Color Image
            pass
            # raise IOError("Palette-color Image. Only grayscale images supported")
        elif interpretation > 2:
            # Palette Color Image
            raise IOError("Only grayscale images supported")

        nRows = info["nRows"]
        nColumns = info["nColumns"]
        nBits = info["nBits"]
        colormap = info["colormap"]
        sampleFormat = info["sampleFormat"]

        if rowMin is None:
            rowMin = 0

        if rowMax is None:
            rowMax = nRows - 1

        if rowMin < 0:
            rowMin = nRows - rowMin

        if rowMax < 0:
            rowMax = nRows - rowMax

        if rowMax < rowMin:
            txt = "Max Row smaller than Min Row. Reverse selection not supported"
            raise NotImplementedError(txt)

        if rowMin >= nRows:
            raise IndexError("Image only has %d rows" % nRows)

        if rowMax >= nRows:
            raise IndexError("Image only has %d rows" % nRows)

        if sampleFormat == SAMPLE_FORMAT_FLOAT:
            if nBits == 32:
                dtype = numpy.float32
            elif nBits == 64:
                dtype = numpy.float64
            else:
                raise ValueError("Unsupported number of bits for a float: %d" % nBits)
        elif sampleFormat in [SAMPLE_FORMAT_UINT, SAMPLE_FORMAT_VOID]:
            if nBits in [8, (8, 8, 8), [8, 8, 8]]:
                dtype = numpy.uint8
            elif nBits in [16, (16, 16, 16), [16, 16, 16]]:
                dtype = numpy.uint16
            elif nBits in [32, (32, 32, 32), [32, 32, 32]]:
                dtype = numpy.uint32
            elif nBits in [64, (64, 64, 64), [64, 64, 64]]:
                dtype = numpy.uint64
            else:
                raise ValueError("Unsupported number of bits for unsigned int: %s" % (nBits,))
        elif sampleFormat == SAMPLE_FORMAT_INT:
            if nBits in [8, (8, 8, 8), [8, 8, 8]]:
                dtype = numpy.int8
            elif nBits in [16, (16, 16, 16), [16, 16, 16]]:
                dtype = numpy.int16
            elif nBits in [32, (32, 32, 32), [32, 32, 32]]:
                dtype = numpy.int32
            elif nBits in [64, (64, 64, 64), [64, 64, 64]]:
                dtype = numpy.int64
            else:
                raise ValueError("Unsupported number of bits for signed int: %s" % (nBits,))
        else:
            raise ValueError("Unsupported combination. Bits = %s  Format = %d" % (nBits, sampleFormat))
        if hasattr(nBits, 'index'):
            image = numpy.zeros((nRows, nColumns, len(nBits)), dtype=dtype)
        elif colormap is not None and interpretation == 3:
            # should I use colormap dtype?
            image = numpy.zeros((nRows, nColumns, 3), dtype=dtype)
        else:
            image = numpy.zeros((nRows, nColumns), dtype=dtype)

        fd = self.fd
        stripOffsets = info["stripOffsets"]  # This contains the file offsets to the data positions
        rowsPerStrip = info["rowsPerStrip"]
        stripByteCounts = info["stripByteCounts"]  # bytes in strip since I do not support compression

        rowStart = 0
        if len(stripOffsets) == 1:
            bytesPerRow = int(stripByteCounts[0] / rowsPerStrip)
            nBytes = stripByteCounts[0]
            if nRows == rowsPerStrip:
                actualBytesPerRow = int(image.nbytes / nRows)
                if actualBytesPerRow != bytesPerRow:
                    logger.warning("Bogus StripByteCounts information")
                    bytesPerRow = actualBytesPerRow
                    nBytes = (rowMax - rowMin + 1) * bytesPerRow
            fd.seek(stripOffsets[0] + rowMin * bytesPerRow)
            readout = numpy.frombuffer(fd.read(nBytes), dtype).copy()
            if self._swap:
                readout.byteswap(True)
            if hasattr(nBits, 'index'):
                readout.shape = -1, nColumns, len(nBits)
            elif info['colormap'] is not None and interpretation == 3:
                readout = colormap[readout]
                readout.shape = -1, nColumns, 3
            else:
                readout.shape = -1, nColumns
            image[...] = readout
        else:
            for i in range(len(stripOffsets)):
                # the amount of rows
                nRowsToRead = rowsPerStrip
                rowEnd = int(min(rowStart + nRowsToRead, nRows))
                if rowEnd < rowMin:
                    rowStart += nRowsToRead
                    continue
                if (rowStart > rowMax):
                    break
                # we are in position
                fd.seek(stripOffsets[i])
                # the amount of bytes to read
                nBytes = stripByteCounts[i]
                if compression_type == 32773:
                    try:
                        bufferBytes = bytes()
                    except Exception:
                        # python 2.5 ...
                        bufferBytes = ""
                    # packBits
                    readBytes = 0
                    # intermediate buffer
                    tmpBuffer = fd.read(nBytes)
                    while readBytes < nBytes:
                        n = struct.unpack('b', tmpBuffer[readBytes:(readBytes + 1)])[0]
                        readBytes += 1
                        if n >= 0:
                            # should I prevent reading more than the
                            # length of the chain? Let's python raise
                            # the exception...
                            bufferBytes += tmpBuffer[readBytes:
                                                     readBytes + (n + 1)]
                            readBytes += (n + 1)
                        elif n > -128:
                            bufferBytes += (-n + 1) * tmpBuffer[readBytes:(readBytes + 1)]
                            readBytes += 1
                        else:
                            # if read -128 ignore the byte
                            continue
                    readout = numpy.frombuffer(bufferBytes, dtype).copy()
                    if self._swap:
                        readout.byteswap(True)

                    if hasattr(nBits, 'index'):
                        readout.shape = -1, nColumns, len(nBits)
                    elif info['colormap'] is not None:
                        readout = colormap[readout]
                        readout.shape = -1, nColumns, 3
                    else:
                        readout.shape = -1, nColumns
                    image[rowStart:rowEnd,:] = readout
                else:
                    readout = numpy.frombuffer(fd.read(nBytes), dtype).copy()
                    if self._swap:
                        readout.byteswap(True)

                    if hasattr(nBits, 'index'):
                        readout.shape = -1, nColumns, len(nBits)
                    elif colormap is not None:
                        readout = colormap[readout]
                        readout.shape = -1, nColumns, 3
                    else:
                        readout.shape = -1, nColumns
                    image[rowStart:rowEnd,:] = readout
                rowStart += nRowsToRead
        if close:
            self.__makeSureFileIsClosed()

        if len(image.shape) == 3:
            # color image
            if self._forceMonoOutput:
                # color image, convert to monochrome
                image = (image[:,:, 0] * 0.114 +
                         image[:,:, 1] * 0.587 +
                         image[:,:, 2] * 0.299).astype(numpy.float32)

        if (rowMin == 0) and (rowMax == (nRows - 1)):
            self._imageDataCacheIndex.insert(0, nImage)
            self._imageDataCache.insert(0, image)
            if len(self._imageDataCacheIndex) > self._maxImageCacheLength:
                self._imageDataCacheIndex = self._imageDataCacheIndex[:self._maxImageCacheLength]
                self._imageDataCache = self._imageDataCache[:self._maxImageCacheLength]

        return image

    def writeImage(self, image0, info=None, software=None, date=None):
        if software is None:
            software = 'PyMca.TiffIO'
        # if date is None:
        #    date = time.ctime()

        self.__makeSureFileIsOpen()
        fd = self.fd
        # prior to do anything, perform some tests
        if not len(image0.shape):
            raise ValueError("Empty image")
        if len(image0.shape) == 1:
            # get a different view
            image = image0[:]
            image.shape = 1, -1
        else:
            image = image0

        if image.dtype == numpy.float64:
            image = image.astype(numpy.float32)
        fd.seek(0)
        mode = fd.mode
        name = fd.name
        if 'w' in mode:
            # we have to overwrite the file
            self.__makeSureFileIsClosed()
            fd = None
            if os.path.exists(name):
                os.remove(name)
            fd = open(name, mode='wb+')
            self._initEmptyFile(fd)
        self.fd = fd

        # read the file size
        self.__makeSureFileIsOpen()
        fd = self.fd
        fd.seek(0, os.SEEK_END)
        endOfFile = fd.tell()
        if fd.tell() == 0:
            self._initEmptyFile(fd)
            fd.seek(0, os.SEEK_END)
            endOfFile = fd.tell()

        # init internal variables
        self._initInternalVariables(fd)
        st = self._structChar

        # get the image file directories
        nImages = self.getImageFileDirectories()
        logger.debug("File contains %d images", nImages)
        if nImages == 0:
            fd.seek(4)
            fmt = st + 'I'
            fd.write(struct.pack(fmt, endOfFile))
        else:
            fd.seek(self._IFD[-1])
            fmt = st + 'H'
            numberOfDirectoryEntries = struct.unpack(fmt, fd.read(struct.calcsize(fmt)))[0]
            fmt = st + 'I'
            pos = self._IFD[-1] + 2 + 12 * numberOfDirectoryEntries
            fd.seek(pos)
            fmt = st + 'I'
            fd.write(struct.pack(fmt, endOfFile))
        fd.flush()

        # and we can write at the end of the file, find out the file length
        fd.seek(0, os.SEEK_END)

        # get the description information from the input information
        if info is None:
            description = info
        else:
            description = "%s" % ""
            for key in info.keys():
                description += "%s=%s\n" % (key, info[key])

        # get the image file directory
        outputIFD = self._getOutputIFD(image, description=description,
                                       software=software,
                                       date=date)

        # write the new IFD
        fd.write(outputIFD)

        # write the image
        if self._swap:
            fd.write(image.byteswap().tobytes())
        else:
            fd.write(image.tobytes())

        fd.flush()
        self.fd = fd
        self.__makeSureFileIsClosed()

    def _initEmptyFile(self, fd=None):
        if fd is None:
            fd = self.fd
        if sys.byteorder == "little":
            order = "II"
            # intel, little endian
            fileOrder = "little"
            self._structChar = '<'
        else:
            order = "MM"
            # motorola, high endian
            fileOrder = "big"
            self._structChar = '>'
        st = self._structChar
        if fileOrder == sys.byteorder:
            self._swap = False
        else:
            self._swap = True
        fd.seek(0)
        fd.write(struct.pack(st + '2s', bytes(order, 'utf-8')))
        fd.write(struct.pack(st + 'H', 42))
        fd.write(struct.pack(st + 'I', 0))
        fd.flush()

    def _getOutputIFD(self, image, description=None, software=None, date=None):
        # the tags have to be in order
        # the very minimum is
        # 256:"NumberOfColumns",           # S or L ImageWidth
        # 257:"NumberOfRows",              # S or L ImageHeight
        # 258:"BitsPerSample",             # S Number of bits per component
        # 259:"Compression",               # SHORT (1 - NoCompression, ...
        # 262:"PhotometricInterpretation", # SHORT (0 - WhiteIsZero, 1 -BlackIsZero, 2 - RGB, 3 - Palette color
        # 270:"ImageDescription",          # ASCII
        # 273:"StripOffsets",              # S or L, for each strip, the byte offset of the strip
        # 277:"SamplesPerPixel",           # SHORT (>=3) only for RGB images
        # 278:"RowsPerStrip",              # S or L, number of rows in each back may be not for the last
        # 279:"StripByteCounts",           # S or L, The number of bytes in the strip AFTER any compression
        # 305:"Software",                  # ASCII
        # 306:"Date",                      # ASCII
        # 339:"SampleFormat",              # SHORT Interpretation of data in each pixel

        nDirectoryEntries = 9
        imageDescription = None
        if description is not None:
            descriptionLength = len(description)
            while descriptionLength < 4:
                description = description + " "
                descriptionLength = len(description)

            if isinstance(description, str):
                raw = description.encode('utf-8')
            else:
                raw = bytes(description, 'utf-8')
            imageDescription = struct.pack("%ds" % len(raw), raw)
            nDirectoryEntries += 1

        # software
        if software is not None:
            softwareLength = len(software)
            while softwareLength < 4:
                software = software + " "
                softwareLength = len(software)
            software = bytes(software, 'utf-8')
            softwarePackedString = struct.pack("%ds" % softwareLength, software)
            nDirectoryEntries += 1
        else:
            softwareLength = 0

        if date is not None:
            dateLength = len(date)
            date = bytes(date, 'utf-8')
            datePackedString = struct.pack("%ds" % dateLength, date)
            dateLength = len(datePackedString)
            nDirectoryEntries += 1
        else:
            dateLength = 0

        if len(image.shape) == 2:
            nRows, nColumns = image.shape
            nChannels = 1
        elif len(image.shape) == 3:
            nRows, nColumns, nChannels = image.shape
        else:
            raise RuntimeError("Image does not have the right shape")
        dtype = image.dtype
        bitsPerSample = int(dtype.str[-1]) * 8

        # only uncompressed data
        compression = 1

        # interpretation, black is zero
        if nChannels == 1:
            interpretation = 1
            bitsPerSampleLength = 0
        elif nChannels == 3:
            interpretation = 2
            bitsPerSampleLength = 3 * 2  # To store 3 shorts
            nDirectoryEntries += 1  # For SamplesPerPixel
        else:
            raise RuntimeError(
                "Image with %d color channel(s) not supported" % nChannels)

        # image description
        if imageDescription is not None:
            descriptionLength = len(imageDescription)
        else:
            descriptionLength = 0

        # strip offsets
        # we are putting them after the directory and the directory is
        # at the end of the file
        self.fd.seek(0, os.SEEK_END)
        endOfFile = self.fd.tell()
        if endOfFile == 0:
            # empty file
            endOfFile = 8

        # rows per strip
        if ALLOW_MULTIPLE_STRIPS:
            # try to segment the image in several pieces
            if not (nRows % 4):
                rowsPerStrip = int(nRows / 4)
            elif not (nRows % 10):
                rowsPerStrip = int(nRows / 10)
            elif not (nRows % 8):
                rowsPerStrip = int(nRows / 8)
            elif not (nRows % 4):
                rowsPerStrip = int(nRows / 4)
            elif not (nRows % 2):
                rowsPerStrip = int(nRows / 2)
            else:
                rowsPerStrip = nRows
        else:
            rowsPerStrip = nRows

        # stripByteCounts
        stripByteCounts = int(nColumns * rowsPerStrip *
                              bitsPerSample * nChannels / 8)

        if descriptionLength > 4:
            stripOffsets0 = (endOfFile + dateLength + descriptionLength +
                             2 + 12 * nDirectoryEntries + 4)
        else:
            stripOffsets0 = (endOfFile + dateLength +
                             2 + 12 * nDirectoryEntries + 4)

        if softwareLength > 4:
            stripOffsets0 += softwareLength

        stripOffsets0 += bitsPerSampleLength

        stripOffsets = [stripOffsets0]
        stripOffsetsLength = 0
        stripOffsetsString = None

        st = self._structChar

        if rowsPerStrip != nRows:
            nStripOffsets = int(nRows / rowsPerStrip)
            fmt = st + 'I'
            stripOffsetsLength = struct.calcsize(fmt) * nStripOffsets
            stripOffsets0 += stripOffsetsLength
            # the length for the stripByteCounts will be the same
            stripOffsets0 += stripOffsetsLength
            stripOffsets = []
            for i in range(nStripOffsets):
                value = stripOffsets0 + i * stripByteCounts
                stripOffsets.append(value)
                if i == 0:
                    stripOffsetsString = struct.pack(fmt, value)
                    stripByteCountsString = struct.pack(fmt, stripByteCounts)
                else:
                    stripOffsetsString += struct.pack(fmt, value)
                    stripByteCountsString += struct.pack(fmt, stripByteCounts)

        logger.debug("IMAGE WILL START AT %d", stripOffsets[0])

        # sample format
        if dtype in [numpy.float32, numpy.float64] or\
           dtype.str[-2] == 'f':
            sampleFormat = SAMPLE_FORMAT_FLOAT
        elif dtype in [numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64]:
            sampleFormat = SAMPLE_FORMAT_UINT
        elif dtype in [numpy.int8, numpy.int16, numpy.int32, numpy.int64]:
            sampleFormat = SAMPLE_FORMAT_INT
        else:
            raise ValueError("Unsupported data type %s" % dtype)

        info = {}
        info["nColumns"] = nColumns
        info["nRows"] = nRows
        info["nBits"] = bitsPerSample
        info["compression"] = compression
        info["photometricInterpretation"] = interpretation
        info["stripOffsets"] = stripOffsets
        if interpretation == 2:
            info["samplesPerPixel"] = 3  # No support for extra samples
        info["rowsPerStrip"] = rowsPerStrip
        info["stripByteCounts"] = stripByteCounts
        info["date"] = date
        info["sampleFormat"] = sampleFormat

        outputIFD = eval('b""')

        fmt = st + "H"
        outputIFD += struct.pack(fmt, nDirectoryEntries)

        fmt = st + "HHII"
        outputIFD += struct.pack(fmt, TAG_NUMBER_OF_COLUMNS,
                                 FIELD_TYPE_OUT['I'], 1, info["nColumns"])
        outputIFD += struct.pack(fmt, TAG_NUMBER_OF_ROWS,
                                 FIELD_TYPE_OUT['I'], 1, info["nRows"])

        if info["photometricInterpretation"] == 1:
            fmt = st + 'HHIHH'
            outputIFD += struct.pack(fmt, TAG_BITS_PER_SAMPLE,
                                     FIELD_TYPE_OUT['H'], 1, info["nBits"], 0)
        elif info["photometricInterpretation"] == 2:
            fmt = st + 'HHII'
            outputIFD += struct.pack(fmt, TAG_BITS_PER_SAMPLE,
                                     FIELD_TYPE_OUT['H'], 3,
                                     info["stripOffsets"][0] - 2 * stripOffsetsLength - descriptionLength -
                                     dateLength - softwareLength - bitsPerSampleLength)
        else:
            raise RuntimeError("Unsupported photometric interpretation")

        fmt = st + 'HHIHH'
        outputIFD += struct.pack(fmt, TAG_COMPRESSION,
                                 FIELD_TYPE_OUT['H'], 1, info["compression"], 0)
        fmt = st + 'HHIHH'
        outputIFD += struct.pack(fmt, TAG_PHOTOMETRIC_INTERPRETATION,
                                 FIELD_TYPE_OUT['H'], 1, info["photometricInterpretation"], 0)
        if imageDescription is not None:
            descriptionLength = len(imageDescription)
            if descriptionLength > 4:
                fmt = st + 'HHII'
                outputIFD += struct.pack(fmt, TAG_IMAGE_DESCRIPTION,
                                         FIELD_TYPE_OUT['s'], descriptionLength,
                                         info["stripOffsets"][0] - 2 * stripOffsetsLength -
                                         descriptionLength)
            else:
                # it has to have length 4
                fmt = st + 'HHI%ds' % descriptionLength
                outputIFD += struct.pack(fmt, TAG_IMAGE_DESCRIPTION,
                                         FIELD_TYPE_OUT['s'],
                                         descriptionLength,
                                         imageDescription)

        if len(stripOffsets) == 1:
            fmt = st + 'HHII'
            outputIFD += struct.pack(fmt, TAG_STRIP_OFFSETS,
                                     FIELD_TYPE_OUT['I'], 1,
                                     info["stripOffsets"][0])
        else:
            fmt = st + 'HHII'
            outputIFD += struct.pack(fmt, TAG_STRIP_OFFSETS,
                                     FIELD_TYPE_OUT['I'],
                                     len(stripOffsets),
                                     info["stripOffsets"][0] - 2 * stripOffsetsLength)

        if info["photometricInterpretation"] == 2:
            fmt = st + 'HHIHH'
            outputIFD += struct.pack(fmt, TAG_SAMPLES_PER_PIXEL,
                                     FIELD_TYPE_OUT['H'], 1,
                                     info["samplesPerPixel"], 0)

        fmt = st + 'HHII'
        outputIFD += struct.pack(fmt, TAG_ROWS_PER_STRIP,
                                 FIELD_TYPE_OUT['I'], 1,
                                 info["rowsPerStrip"])

        if len(stripOffsets) == 1:
            fmt = st + 'HHII'
            outputIFD += struct.pack(fmt, TAG_STRIP_BYTE_COUNTS,
                                     FIELD_TYPE_OUT['I'], 1,
                                     info["stripByteCounts"])
        else:
            fmt = st + 'HHII'
            outputIFD += struct.pack(fmt, TAG_STRIP_BYTE_COUNTS,
                                     FIELD_TYPE_OUT['I'], len(stripOffsets),
                                     info["stripOffsets"][0] - stripOffsetsLength)

        if software is not None:
            if softwareLength > 4:
                fmt = st + 'HHII'
                outputIFD += struct.pack(fmt, TAG_SOFTWARE,
                                         FIELD_TYPE_OUT['s'],
                                         softwareLength,
                                         info["stripOffsets"][0] -
                                         2 * stripOffsetsLength -
                                         descriptionLength - softwareLength - dateLength)
            else:
                # it has to have length 4
                fmt = st + 'HHI%ds' % softwareLength
                outputIFD += struct.pack(fmt, TAG_SOFTWARE,
                                         FIELD_TYPE_OUT['s'],
                                         softwareLength,
                                         softwarePackedString)

        if date is not None:
            fmt = st + 'HHII'
            outputIFD += struct.pack(fmt, TAG_DATE,
                                     FIELD_TYPE_OUT['s'],
                                     dateLength,
                                     info["stripOffsets"][0] -
                                     2 * stripOffsetsLength -
                                     descriptionLength - dateLength)

        fmt = st + 'HHIHH'
        outputIFD += struct.pack(fmt, TAG_SAMPLE_FORMAT,
                                 FIELD_TYPE_OUT['H'], 1,
                                 info["sampleFormat"], 0)
        fmt = st + 'I'
        outputIFD += struct.pack(fmt, 0)

        if info["photometricInterpretation"] == 2:
            outputIFD += struct.pack('HHH', info["nBits"],
                                     info["nBits"], info["nBits"])

        if softwareLength > 4:
            outputIFD += softwarePackedString

        if date is not None:
            outputIFD += datePackedString

        if imageDescription is not None:
            if descriptionLength > 4:
                outputIFD += imageDescription

        if stripOffsetsString is not None:
            outputIFD += stripOffsetsString
            outputIFD += stripByteCountsString

        return outputIFD



#!/usr/bin/python
# coding: utf-8
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Benchmark for file reading"""

__author__ = "Jérôme Kieffer"
__date__ = "03/04/2020"
__license__ = "MIT"
__copyright__ = "2016-2020 European Synchrotron Radiation Facility, Grenoble, France"

import json
import sys
import time
import timeit
import os
import platform
import subprocess
import numpy
import fabio
import os.path as op
import logging

# To use use the locally build version of PyFAI, use ../bootstrap.py
try:
    from .. import open as fabio_open, version, date
except ImportError:
    from fabio import open as fabio_open, version, date
from ..test import utilstest

datasets = ["mb_LP_1_001.img",
            "Cr8F8140k103.0026",
            "run2_1_00148.cbf",
            "F2K_Seb_Lyso0675.edf",
            "fit2d_click.msk",
            "GE_aSI_detector_image_1529",
            "i01f0001.kcd",
            "example.mar2300",
            "corkcont2_H_0089.mccd",
            "b191_1_9_1.img",
            "image0001.pgm",
            "mgzn-20hpt.img",
            "oPPA_5grains_0001.tif",
            "XSDataImage.xml", ]

setup = """
import fabio
"""
stmt = "data = fabio.open(r'%s').data"


def run_benchmark(number=10, repeat=3):
    """
    :param number: Measure timimg over number of executions
    :param repeat: number of measurement, takes the best of them

    """
    print("Averaging over %i repetitions (best of %s)." % (number, repeat))
    print("Python %s" % sys.version)
    print("FabIO %s (%s)" % (version, date))
    print("#" * 80)
    print("     Module           filename        \t file size \t image size \t read time (ms) \t ms/Mpix")
    for img in datasets:
        fn = utilstest.UtilsTest.getimage(img)
        fimg = fabio_open(fn)
        file_size = os.stat(fn).st_size / 1.0e6  # MB
        img_size = fimg.data.size / 1.0e6  # Mpix
        timer = timeit.Timer(stmt % fn, setup + (stmt % fn))
        tmin = min([i / (0.001 * number) for i in timer.repeat(repeat=repeat, number=number)])
        print("%13s %25s %.3f Mb \t %.3f Mpix \t  %.3f ms \t %.3f ms/Mpix" %
              (fimg.__class__.__name__, img, file_size, img_size, tmin, tmin / img_size))


run = run_benchmark



# coding: utf-8
#
#    Project: X-ray image reader
#             https://github.com/silx-kit/fabio
#
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE

"""
Authors: Jerome Kieffer, ESRF
         email:jerome.kieffer@esrf.fr

kcd images are 2D images written by the old KappaCCD diffractometer built by Nonius in the 1990's
Based on the edfimage.py parser.
"""

import numpy
import logging
import os

import string
from .fabioimage import FabioImage
logger = logging.getLogger(__name__)

import io
if not hasattr(io, "SEEK_END"):
    SEEK_END = 2
else:
    SEEK_END = io.SEEK_END

DATA_TYPES = {"u16": numpy.uint16}

MINIMUM_KEYS = [
    # 'ByteOrder', Assume little by default
    'Data type',
    'X dimension',
    'Y dimension',
    'Number of readouts']

DEFAULT_VALUES = {"Data type": "u16"}

ALPHANUM = bytes(string.digits + string.ascii_letters + ". ", encoding="ASCII")


class KcdImage(FabioImage):
    """
    Read the Nonius kcd data format """

    DESCRIPTION = "KCD file format from Nonius's KappaCCD diffractometer"

    DEFAULT_EXTENSIONS = ["kcd"]

    def _readheader(self, infile):
        """
        Read in a header in some KCD format from an already open file
        """
        one_line = infile.readline()

        asciiHeader = True
        for oneChar in one_line.strip():
            if oneChar not in ALPHANUM:
                asciiHeader = False

        if asciiHeader is False:
            # This does not look like an KappaCCD file
            logger.warning("First line of %s does not seam to be ascii text!" % infile.name)
        end_of_headers = False
        while not end_of_headers:
            one_line = infile.readline()
            try:
                one_line = one_line.decode("ASCII")
            except UnicodeDecodeError:
                end_of_headers = True
            else:
                if len(one_line) > 100:
                    end_of_headers = True
            if not end_of_headers:
                if one_line.strip() == "Binned mode":
                    one_line = "Mode = Binned"
                if "=" in one_line:
                    key, val = one_line.split('=', 1)
                    key = key.strip()
                    self.header[key] = val.strip()
                else:
                    end_of_headers = True

        missing = []
        for item in MINIMUM_KEYS:
            if item not in self.header:
                missing.append(item)
        if len(missing) > 0:
            logger.debug("KCD file misses the keys " + " ".join(missing))

    def read(self, fname, frame=None):
        """
        Read in header into self.header and
            the data   into self.data
        """
        self.header = self.check_header()
        self.resetvals()
        with self._open(fname, "rb") as infile:
            self._readheader(infile)
            # Compute image size
            try:
                dim1 = int(self.header['X dimension'])
                dim2 = int(self.header['Y dimension'])
                self._shape = dim2, dim1
            except (KeyError, ValueError):
                raise IOError("KCD file %s is corrupt, cannot read it" % fname)
            try:
                bytecode = DATA_TYPES[self.header['Data type']]
            except KeyError:
                bytecode = numpy.uint16
                logger.warning("Defaulting type to uint16")
            self._dtype = numpy.dtype(bytecode)
            try:
                nbReadOut = int(self.header['Number of readouts'])
            except KeyError:
                logger.warning("Defaulting number of ReadOut to 1")
                nbReadOut = 1
            expected_size = dim1 * dim2 * self._dtype.itemsize * nbReadOut

            try:
                infile.seek(-expected_size, SEEK_END)
            except Exception:
                logger.debug("Backtrace", exc_info=True)
                logger.warning("seeking from end is not implemeneted for file %s", fname)
                if hasattr(infile, "measure_size"):
                    fileSize = infile.measure_size()
                elif hasattr(infile, "size"):
                    fileSize = infile.size
                elif hasattr(infile, "getSize"):
                    fileSize = infile.getSize()
                else:
                    logger.warning("Unable to guess the file-size of %s", fname)
                    fileSize = os.stat(fname)[6]
                infile.seek(fileSize - expected_size - infile.tell(), 1)
            block = infile.read(expected_size)
        # infile.close()

        # now read the data into the array
        self.data = numpy.zeros((dim2, dim1), numpy.int32)
        stop = 0
        for i in range(nbReadOut):
            start = stop
            stop = (i + 1) * expected_size // nbReadOut
            data = numpy.frombuffer(block[start: stop], self._dtype).copy()
            data.shape = dim2, dim1
            if not numpy.little_endian:
                data.byteswap(True)
            self.data += data
        self._dtype = None
        self._shape = None
        self.resetvals()
        return self

    @staticmethod
    def checkData(data=None):
        if data is None:
            return None
        else:
            return data.astype(int)


kcdimage = KcdImage



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Fable Input Output
#             https://github.com/silx-kit/fabio
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Test cases for filename deconstruction


testsuite by Jerome Kieffer (Jerome.Kieffer@esrf.eu)
28/11/2014
"""

import unittest
import os
import logging
import tempfile


logger = logging.getLogger(__name__)

import fabio
import numpy

CASES = [
    (1, 'edf', "data0001.edf"),
    (10001, 'edf', "data10001.edf"),
    (10001, 'edf', "data10001.edf.gz"),
    (10001, 'edf', "data10001.edf.bz2"),
    (2, 'marccd', "data0002.mccd"),
    (12345, 'marccd', "data12345.mccd"),
    (10001, 'marccd', "data10001.mccd.gz"),
    (10001, 'marccd', "data10001.mccd.bz2"),
    (123, 'marccd', "data123.mccd.gz"),
    (3, 'tif_or_pilatus', "data0003.tif"),
    (4, 'tif_or_pilatus', "data0004.tiff"),
    (12, 'bruker', "sucrose101.012.gz"),
    (99, 'bruker', "sucrose101.099"),
    (99, 'bruker', "sucrose101.0099"),
    (99, 'bruker', "sucrose101.0099.bz2"),
    (99, 'bruker', "sucrose101.0099.gz"),
    (2, 'fit2dmask', "fit2d.msk"),
    (None, 'fit2dmask', "mymask.msk"),
    (670005, 'edf', 'S82P670005.edf'),
    (670005, 'edf', 'S82P670005.edf.gz'),
    # based on only the name it can be either img or oxd
    (1, 'dtrek_or_oxd_or_hipic_or_raxis', 'mb_LP_1_001.img'),
    (2, 'dtrek_or_oxd_or_hipic_or_raxis', 'mb_LP_1_002.img.gz'),
    (3, 'dtrek_or_oxd_or_hipic_or_raxis', 'mb_LP_1_003.img.bz2'),
    (3, 'dtrek_or_oxd_or_hipic_or_raxis', os.path.join("data", 'mb_LP_1_003.img.bz2')),
]

MORE_CASES = [
    ("data0010.edf", "data0012.edf", 10),
    ("data1000.pnm", "data999.pnm", 1000),
    ("data0999.pnm", "data1000.pnm", 999),
    ("data123457.edf", "data123456.edf", 123457),
    ("d0ata000100.mccd", "d0ata000012.mccd", 100),
    (os.path.join("images/sampledir", "P33S670003.edf"),
     os.path.join("images/sampledir", "P33S670002.edf"), 670003),
    (os.path.join("images/P33S67", "P33S670003.edf"),
     os.path.join("images/P33S67", "P33S670002.edf"), 670003),
    ("image2301.mar2300", "image2300.mar2300", 2301),
    ("image2300.mar2300", "image2301.mar2300", 2300),
    ("image.0123", "image.1234", 123),
    ("mymask.msk", "mymask.msk", None),
    ("data_123.mccd.bz2", "data_001.mccd.bz2", 123)
]


class TestFilenames(unittest.TestCase):
    """ check the name -> number, type conversions """

    def test_many_cases(self):
        """ loop over CASES """
        for num, typ, name in CASES:
            obj = fabio.FilenameObject(filename=name)
            self.assertEqual(num, obj.num, name + " num=" + str(num) +
                             " != obj.num=" + str(obj.num))
            self.assertEqual(typ, "_or_".join(obj.format),
                             name + " " + "_or_".join(obj.format))
            self.assertEqual(name, obj.tostring(), name + " " + obj.tostring())

    def test_more_cases(self):
        for nname, oname, num in MORE_CASES:
            name = fabio.construct_filename(oname, num)
            self.assertEqual(name, nname)

    def test_more_cases_jump(self):
        for nname, oname, num in MORE_CASES:
            name = fabio.jump_filename(oname, num)
            self.assertEqual(name, nname)


class TestFilenameObjects(unittest.TestCase):

    def setUp(self):
        """ make a small test dataset """
        self.datashape = (10,11)
        self.nframes = 5
        self.tempdir = tempfile.mkdtemp()
        self.fnames = [os.path.join(self.tempdir, "FNO%04d.edf" % iframe)
                       for iframe in range(self.nframes)]
        data = numpy.zeros(self.datashape,numpy.uint16)
        im = fabio.edfimage.edfimage(data)
        for iframe, fname in enumerate(self.fnames):
            im.header["checkthing"] = str(iframe)
            im.write(fname)

    def tearDown(self):
        for name in self.fnames:
            os.remove(name)
        os.rmdir(self.tempdir)

    def test_files_are_being_opened(self):
        """Regression test for Fable"""
        for iframe, fname in enumerate(self.fnames):
            obj = fabio.FilenameObject(filename=fname)
            # read via the FilenameObject
            o1  = fabio.open(obj)
            self.assertEqual(o1.data.shape, self.datashape)
            self.assertEqual(o1.header['checkthing'], str(iframe))
            # And via the tostring
            o2  = fabio.open(obj.tostring())
            self.assertEqual(o2.data.shape, self.datashape)
            self.assertEqual(o2.header['checkthing'], str(iframe))

    def test_FileNameObject_can_iterate(self):
        """Regression test for Fable"""
        obj = fabio.FilenameObject(filename=self.fnames[0])
        for iframe, fname in enumerate(self.fnames):
            obj.num = iframe
            # read via the FilenameObject
            o1  = fabio.open(obj)
            self.assertEqual(o1.data.shape, self.datashape)
            self.assertEqual(o1.header['checkthing'], str(iframe))
            # And via the tostring
            o2  = fabio.open(obj.tostring())
            self.assertEqual(o2.data.shape, self.datashape)
            self.assertEqual(o2.header['checkthing'], str(iframe))
            # And the real name
            o3  = fabio.open(fname)
            self.assertEqual(o3.data.shape, self.datashape)
            self.assertEqual(o3.header['checkthing'], str(iframe))


def suite():
    loadTests = unittest.defaultTestLoader.loadTestsFromTestCase
    testsuite = unittest.TestSuite()
    testsuite.addTest(loadTests(TestFilenames))
    testsuite.addTest(loadTests(TestFilenameObjects))
    return testsuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())
